"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import json
import logging
import os
from collections.abc import Mapping
from typing import Optional

import pytest

import inmanta.resources
from inmanta import config, const, module
from inmanta.agent.code_manager import CodeManager
from inmanta.ast import CompilerException, ExternalException, RuntimeException
from inmanta.const import ResourceState
from inmanta.data import Environment, Resource
from inmanta.export import DependencyCycleException
from inmanta.module import InmantaModuleRequirement
from inmanta.server import SLICE_RESOURCE
from inmanta.server.server import Server
from packaging.version import Version
from utils import LogSequence, PipIndex, create_python_package


async def assert_resource_set_assignment(environment, assignment: dict[str, Optional[str]]) -> None:
    """
    Verify whether the resources on the server are assigned to the resource sets given via the assignment argument.

    :param environment: environment uuid
    :param assignment: Map the value of name attribute of resource Res to the resource set that resource is expected to
                       belong to.
    """
    resources = await Resource.get_resources_in_latest_version(environment=environment)
    assert len(resources) == len(assignment)
    actual_assignment = {r.attributes["key"]: r.resource_set for r in resources}
    assert actual_assignment == assignment


def test_attribute_mapping_export(snippetcompiler):
    """
    Verify Python-defined attribute mapping between model resource and exported resource.
    """
    snippetcompiler.setup_for_snippet(
        """\
        import exp

        exp::Test(name="a", agent="x")
        exp::Test(name="b", agent="x", send_event=false, receive_events=true)
        exp::Test(name="c", agent="x", send_event=true, receive_events=false)
        """,
        autostd=True,
        project_requires=[module.InmantaModuleRequirement.parse("std>=6.1")],
    )

    _version, json_value = snippetcompiler.do_export()

    assert len(json_value) == 3
    resources: Mapping[str, Mapping[str, object]] = {
        resource.name: {
            # base resource mapping
            const.RESOURCE_ATTRIBUTE_SEND_EVENTS: getattr(resource, const.RESOURCE_ATTRIBUTE_SEND_EVENTS),
            const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS: getattr(resource, const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS),
            # custom mapping
            "mapped": resource.mapped,
        }
        for resource in json_value.values()
    }
    assert resources == {
        "a": {
            "mapped": "mapped_value_a",
            const.RESOURCE_ATTRIBUTE_SEND_EVENTS: True,
            const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS: True,
        },
        "b": {
            "mapped": "mapped_value_b",
            const.RESOURCE_ATTRIBUTE_SEND_EVENTS: False,
            const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS: True,
        },
        "c": {
            "mapped": "mapped_value_c",
            const.RESOURCE_ATTRIBUTE_SEND_EVENTS: True,
            const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS: False,
        },
    }


def test_id_mapping_export(snippetcompiler):
    """
    Verify id mapping for a virtual id value, i.e. not a field of the resource.
    """
    snippetcompiler.setup_for_snippet(
        """import exp

        exp::Test(name="a", agent="b")
        """,
        autostd=True,
    )

    _version, json_value = snippetcompiler.do_export()

    assert len(json_value) == 1
    resource = list(json_value.values())[0]
    assert not hasattr(resource, resource.id.attribute)
    assert resource.id.attribute_value == "test_value_a"


def test_unknown_agent(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """import exp
        import tests

        exp::Test(name="a", agent=tests::unknown())
        """,
        autostd=True,
    )
    _version, json_value = snippetcompiler.do_export()

    assert len(json_value) == 0


def test_unknown_attribute_value(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """import exp
        import tests

        exp::Test(name=tests::unknown(), agent="b")
        """,
        autostd=True,
    )
    _version, json_value = snippetcompiler.do_export()

    assert len(json_value) == 0


def test_ignore_resource(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """import exp
        import tests

        exp::Test(name="a", agent="b", managed=false)
        """,
        autostd=True,
    )
    _version, json_value = snippetcompiler.do_export()

    assert len(json_value) == 0


def test_ignore_resource_requires(snippetcompiler, caplog):
    snippetcompiler.setup_for_snippet(
        """import exp
        import tests

        a = exp::Test(name="a", agent="aa", managed=false)
        b = exp::Test(name="b", agent="aa", requires=a)
        c = exp::Test(name="c", agent="aa", requires=b)
        """,
        autostd=True,
    )
    _version, json_value = snippetcompiler.do_export()
    assert len(json_value) == 2
    assert_count = 0
    for resource_id, resource in json_value.items():
        if resource_id.attribute_value == "test_value_b":
            assert len(resource.requires) == 0
            assert_count += 1

        elif resource_id.attribute_value == "test_value_c":
            assert len(resource.requires) == 1
            assert_count += 1

    warning = [
        x
        for x in caplog.records
        if x.msg == "The resource %s had requirements before flattening, but not after flattening."
        " Initial set was %s. Perhaps provides relation is not wired through correctly?"
    ]
    assert len(warning) == 0
    assert assert_count == 2


def test_unknown_in_id_requires(snippetcompiler, caplog):
    """
    Test to validate that resources that have an unknown in their ID attributes, are removed from requires
    """
    snippetcompiler.setup_for_snippet(
        """import exp
        import tests

        a = exp::Test(name=tests::unknown(), agent="aa")
        b = exp::Test(name="b", agent="aa", requires=a)
        c = exp::Test(name="c", agent="aa", requires=b)
        """,
        autostd=True,
    )
    _version, json_value = snippetcompiler.do_export()

    assert len(json_value) == 2
    assert_count = 0
    for resource_id, resource in json_value.items():
        if resource_id.attribute_value == "test_value_b":
            assert len(resource.requires) == 0
            assert_count += 1

        elif resource_id.attribute_value == "test_value_c":
            assert len(resource.requires) == 1
            assert_count += 1

    warning = [
        x
        for x in caplog.records
        if x.msg == "The resource %s had requirements before flattening, but not after flattening."
        " Initial set was %s. Perhaps provides relation is not wired through correctly?"
    ]
    assert len(warning) == 0
    assert assert_count == 2


def test_unknown_in_attribute_requires(snippetcompiler, caplog):
    """
    Test to validate that resources that have an unknown in their ID attributes, are removed from requires
    """
    snippetcompiler.setup_for_snippet(
        """import exp
        import tests

        a = exp::Test(name="a", agent="aa", field1=tests::unknown())
        b = exp::Test(name="b", agent="aa", requires=a)
        c = exp::Test(name="c", agent="aa", requires=b)
        """,
        autostd=True,
    )
    _version, json_value, status = snippetcompiler.do_export(include_status=True)

    assert len(json_value) == 3
    assert len([x for x in status.values() if x == const.ResourceState.available]) == 2
    assert len([x for x in status.values() if x == const.ResourceState.undefined]) == 1

    warning = [
        x
        for x in caplog.records
        if x.msg == "The resource %s had requirements before flattening, but not after flattening."
        " Initial set was %s. Perhaps provides relation is not wired through correctly?"
    ]
    assert len(warning) == 0


async def test_empty_server_export(snippetcompiler, server, client, environment):
    snippetcompiler.setup_for_snippet(
        """
            h = std::Host(name="test", os=std::linux)
        """,
        extra_index_url=["example.inmanta.com/index"],
        autostd=True,
    )
    await snippetcompiler.do_export_and_deploy()

    response = await client.list_versions(tid=environment)
    assert response.code == 200
    assert len(response.result["versions"]) == 1
    assert response.result["versions"][0]["pip_config"] == {
        "extra-index-url": ["example.inmanta.com/index"],
        "index-url": None,
        "pre": None,
        "use-system-config": False,
    }


async def test_server_export(snippetcompiler, server: Server, client, environment):
    snippetcompiler.setup_for_snippet(
        """
            import std::testing
            f = std::testing::NullResource(name="test")
        """,
        autostd=True,
    )
    await snippetcompiler.do_export_and_deploy()

    result = await client.list_versions(tid=environment)
    assert result.code == 200
    assert len(result.result["versions"]) == 1
    assert result.result["versions"][0]["total"] == 1

    version = result.result["versions"][0]["version"]
    result = await client.get_version(tid=environment, id=result.result["versions"][0]["version"])
    assert result.code == 200

    for res in result.result["resources"]:
        res["attributes"]["id"] = res["id"]
        resource = inmanta.resources.Resource.deserialize(res["attributes"])
        assert resource.version == resource.id.version == version

    resources = await server.get_slice(SLICE_RESOURCE).get_resources_in_latest_version(
        environment=await Environment.get_by_id(environment)
    )

    assert resources[0].attributes["version"] == version


async def test_dict_export_server(snippetcompiler, server, client, environment):
    config.Config.set("config", "environment", environment)
    snippetcompiler.setup_for_snippet(
        """
import exp

a = exp::Test2(mydict={"a":"b"}, mylist=["a","b"])
""",
        autostd=True,
    )

    await snippetcompiler.do_export_and_deploy()

    result = await client.list_versions(tid=environment)
    assert result.code == 200
    assert len(result.result["versions"]) == 1
    assert result.result["versions"][0]["total"] == 1


async def test_project_constraints_at_exporter_boundary(snippetcompiler, server, client, environment, tmpdir, agent):
    """
    Test that constraints set in the requirements.txt of the project are propagated into the agent's EnvBlueprint
    during the export flow.
    """
    index: PipIndex = PipIndex(str(tmpdir.join("index")))
    for v in ("1.0.0", "2.0.0"):
        create_python_package(
            "dependency_package",
            Version(v),
            str(tmpdir.join(f"dependency_package-{v}")),
            publish_index=index,
        )

    config.Config.set("config", "environment", environment)
    snippetcompiler.setup_for_snippet(
        """
import many_dependencies

a = many_dependencies::Test(name="my_test_resource")
""",
        autostd=True,
        index_url=index.url,
        extra_index_url=["https://pypi.org/simple"],
    )

    await snippetcompiler.do_export_and_deploy()

    result = await client.list_versions(tid=environment)
    assert result.code == 200
    assert len(result.result["versions"]) == 1
    assert result.result["versions"][0]["total"] == 1

    codemanager = CodeManager(agent._client)
    module_install_specs, _ = await codemanager.get_code(
        environment=environment,
        version=1,
        resource_types=["many_dependencies::Test"],
    )

    assert len(module_install_specs) > 0
    for module_install_spec in module_install_specs:
        assert module_install_spec.blueprint.project_constraints is None

    constraints = [
        "dependency-package<2.0.0",
        "jinja2<3.1.6",
    ]

    snippetcompiler.setup_for_snippet(
        """
import many_dependencies

a = many_dependencies::Test(name="my_test_resource")
""",
        autostd=True,
        python_requires=[inmanta.util.parse_requirement(requirement=constraint) for constraint in constraints],
        index_url=index.url,
        extra_index_url=["https://pypi.org/simple"],
    )

    await snippetcompiler.do_export_and_deploy()

    result = await client.list_versions(tid=environment)
    assert result.code == 200
    assert len(result.result["versions"]) == 2
    assert result.result["versions"][1]["total"] == 1

    module_install_specs, _ = await codemanager.get_code(
        environment=environment,
        version=2,
        resource_types=["many_dependencies::Test"],
    )
    assert len(module_install_specs) > 0

    for module_install_spec in module_install_specs:
        assert module_install_spec.blueprint.project_constraints == "\n".join(constraints)


async def test_old_compiler(server, client, environment):
    result = await client.put_version(tid=environment, version=123456, resources=[], unknowns=[], version_info={})
    assert result.code == 400


def test_dict_export(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import exp

a = exp::Test2(mydict={"a":"b"}, mylist=["a","b"])
""",
        autostd=True,
    )
    _version, json_value, status = snippetcompiler.do_export(include_status=True)

    assert len(json_value) == 1


def test_export_null_in_collection(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import exp

a = exp::Test2(mydict={"a": null}, mylist=["a",null])
""",
        autostd=True,
    )
    _version, json_value, status = snippetcompiler.do_export(include_status=True)

    assert len(json_value) == 1
    json_dict = snippetcompiler.get_exported_json()
    resource = json_dict[0]
    assert resource["mylist"] == ["a", None]
    assert resource["mydict"] == {"a": None}


def test_export_unknown_in_collection(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import exp
import tests

a = exp::Test2(mydict={"a": tests::unknown()}, mylist=["a"])
b = exp::Test2(name="idb", mydict={"a": "b"}, mylist=["a", tests::unknown()])
""",
        autostd=True,
    )
    _version, json_value, status = snippetcompiler.do_export(include_status=True)

    assert len(json_value) == 2
    assert status["exp::Test2[agenta,name=ida]"] == ResourceState.undefined
    assert status["exp::Test2[agenta,name=idb]"] == ResourceState.undefined


def test_1934_cycle_in_dep_mgmr(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import exp

a = exp::RequiresTest()
b = exp::RequiresTest(name="idb")
a.requires += b
""",
        autostd=True,
    )
    with pytest.raises(DependencyCycleException):
        snippetcompiler.do_export()


def test_bad_value_in_dep_mgmr(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import exp

a = exp::RequiresTest(do_break=1)
""",
        autostd=True,
    )
    with pytest.raises(Exception, match="Invalid id for resource xyz"):
        snippetcompiler.do_export()


def test_bad_value_in_dep_mgmr_2(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import exp

a = exp::RequiresTest(do_break=2)
""",
        autostd=True,
    )
    with pytest.raises(
        Exception,
        match="A dependency manager inserted the object <object object at .*> of type <class 'object'> "
        "into a requires relation. However, only string, Resource or Id are allowable types",
    ):
        snippetcompiler.do_export()


def test_bad_value_in_dep_mgmr_3(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import exp

a = exp::RequiresTest(do_break=3)
""",
        autostd=True,
    )
    with pytest.raises(
        Exception,
        match="A dependency manager inserted a resource id without version this is not allowed aa::Bbbb\\[agent,name=agent\\]",
    ):
        snippetcompiler.do_export()


def test_2121_wrapped_proxy_serialize(snippetcompiler):
    snippetcompiler.setup_for_snippet(
        """
import exp

dct = {"a": 1, "b": 2}

x = exp::WrappedProxyTest(
    name = "my_wrapped_proxy_test",
    agent = "my_agent",
    my_list = [1, 2, 3],
    my_dict = {
        "dct": dct,
    }
)
        """,
        autostd=True,
    )
    snippetcompiler.do_export()
    tmp_file: str = os.path.join(snippetcompiler.project_dir, "dump.json")
    with open(tmp_file) as f:
        export: dict = json.loads(f.read())
        my_dict: dict = {"dct": {"a": 1, "b": 2}}
        assert len(export) == 1
        print(export[0])
        assert export[0]["wrapped_proxies"] == {
            "my_list": [1, 2, 3],
            "my_dict": my_dict,
            "deep_dict": {"multi_level": my_dict},
        }


def test_2121_wrapped_self_serialize(snippetcompiler):
    """
    Make sure DynamicProxies representing an entity are not serialized.
    """
    snippetcompiler.setup_for_snippet(
        """
import exp

exp::WrappedSelfTest(
    name = "my_wrapped_self_test",
    agent = "my_agent",
)
        """,
        autostd=True,
    )
    with pytest.raises(ExternalException) as e:
        snippetcompiler.do_export()
    assert "not JSON serializable" in e.value.format_trace()


def test_3787_key_error_export(snippetcompiler):
    """
    Check the error message of an export with a KeyError
    The Key error happens in get_real_name() of class Test3
    """
    snippetcompiler.setup_for_snippet(
        """
import exp

exp::Test3(
    name="tom",
    names={
        "bob": "alice",
        "alice": "bob",
    },
    agent=std::AgentConfig(
        autostart=true,
        agentname="bob",
        uri="local:",
    ),
)
        """,
        autostd=True,
    )
    with pytest.raises(ExternalException) as e:
        snippetcompiler.do_export()
    assert (
        e.value.format_trace()
        == "Failed to get attribute 'real_name' for export on 'exp::Test3'\ncaused by:\nKeyError: 'tom'\n"
    )


@pytest.mark.parametrize("soft_delete", [True, False])
async def test_resource_set(snippetcompiler, modules_dir: str, environment, client, agent, soft_delete: bool) -> None:
    """
    Test that resource sets are exported correctly, when a full compile or an incremental compile is done.
    """

    async def export_model(
        model: str,
        partial_compile: bool,
        resource_sets_to_remove: Optional[list[str]] = None,
    ) -> None:
        snippetcompiler.setup_for_snippet(
            model,
            extra_index_url=["example.inmanta.com/index"],
            autostd=True,
        )
        await snippetcompiler.do_export_and_deploy(
            partial_compile=partial_compile,
            resource_sets_to_remove=resource_sets_to_remove,
            soft_delete=soft_delete,
        )

    # Full compile
    await export_model(
        model="""
import test_resources

a = test_resources::Resource(value="A", agent="A", key="the_resource_a")
b = test_resources::Resource(value="A", agent="A", key="the_resource_b")
c = test_resources::Resource(value="A", agent="A", key="the_resource_c")
d = test_resources::Resource(value="A", agent="A", key="the_resource_d")
e = test_resources::Resource(value="A", agent="A", key="the_resource_e")
y = test_resources::Resource(value="A", agent="A", key="the_resource_y")
z = test_resources::Resource(value="A", agent="A", key="the_resource_z")
std::ResourceSet(name="resource_set_1", resources=[a,c])
std::ResourceSet(name="resource_set_2", resources=[b])
std::ResourceSet(name="resource_set_3", resources=[d, e])
        """,
        partial_compile=False,
    )
    await assert_resource_set_assignment(
        environment,
        assignment={
            "the_resource_a": "resource_set_1",
            "the_resource_b": "resource_set_2",
            "the_resource_c": "resource_set_1",
            "the_resource_d": "resource_set_3",
            "the_resource_e": "resource_set_3",
            "the_resource_y": None,
            "the_resource_z": None,
        },
    )

    # Partial compile
    await export_model(
        model="""
    import test_resources

    a = test_resources::Resource(value="A", agent="A", key="the_resource_a")
    c2 = test_resources::Resource(value="A", agent="A", key="the_resource_c2")
    f = test_resources::Resource(value="A", agent="A", key="the_resource_f")
    # y is a shared resource, identical to the one in previous compile
    y = test_resources::Resource(value="A", agent="A", key="the_resource_y")
    # z is a shared resource not present in this model
    std::ResourceSet(name="resource_set_1", resources=[a,c2])
    std::ResourceSet(name="resource_set_4", resources=[f])
            """,
        partial_compile=True,
        resource_sets_to_remove=["resource_set_2"],
    )
    await assert_resource_set_assignment(
        environment,
        assignment={
            "the_resource_a": "resource_set_1",
            "the_resource_c2": "resource_set_1",
            "the_resource_d": "resource_set_3",
            "the_resource_e": "resource_set_3",
            "the_resource_f": "resource_set_4",
            "the_resource_y": None,
            "the_resource_z": None,
        },
    )

    # Test soft_delete option

    model = """
        import test_resources

        g = test_resources::Resource(value="A", agent="A", key="the_resource_g")
        std::ResourceSet(name="resource_set_5", resources=[g])

        """
    if not soft_delete:
        with pytest.raises(
            Exception,
            match=(
                "Invalid request: Following resource sets are present in the removed resource"
                " sets and in the resources that are exported: {'resource_set_5'}"
            ),
        ):
            await export_model(
                model=model,
                partial_compile=True,
                resource_sets_to_remove=["resource_set_5"],
            )

    else:
        await export_model(
            model=model,
            partial_compile=True,
            resource_sets_to_remove=["resource_set_5"],
        )
        await assert_resource_set_assignment(
            environment,
            assignment={
                "the_resource_a": "resource_set_1",
                "the_resource_c2": "resource_set_1",
                "the_resource_d": "resource_set_3",
                "the_resource_e": "resource_set_3",
                "the_resource_f": "resource_set_4",
                "the_resource_g": "resource_set_5",  # Check it didn't get removed
                "the_resource_y": None,
                "the_resource_z": None,
            },
        )

    response = await client.list_versions(tid=environment)
    assert response.code == 200

    # One of the 3 partial compiles is expected to fail when soft_delete is true:
    assert len(response.result["versions"]) == 2 + soft_delete
    last_version_nr = 0
    expected_pip_config = {
        "extra-index-url": ["example.inmanta.com/index"],
        "index-url": None,
        "pre": None,
        "use-system-config": False,
    }
    for version in response.result["versions"]:
        assert version["pip_config"] == expected_pip_config, f"failed for version: {version['version']}"
        last_version_nr = version["version"]

    # abusing this setup to test get_pip_config
    response = await agent._client.get_pip_config(tid=environment, version=last_version_nr)
    assert response.code == 200
    assert response.result["data"] == expected_pip_config

    response = await agent._client.get_pip_config(tid=environment, version=500)
    assert response.code == 404


async def test_resource_in_multiple_resource_sets(snippetcompiler, environment) -> None:
    """
    test that an error is raised if a resource is in multiple
    resource_sets
    """
    model = """
import test_resources

a = test_resources::Resource(value="A", agent="A", key="the_resource_a")
std::ResourceSet(name="resource_set_1", resources=[a])
std::ResourceSet(name="resource_set_2", resources=[a])
"""

    snippetcompiler.setup_for_snippet(
        model,
        autostd=True,
    )
    with pytest.raises(CompilerException) as e:
        await snippetcompiler.do_export_and_deploy()
    assert str(e.value).startswith(
        "resource 'test_resources::Resource[A,key=the_resource_a]' can not be part of multiple " "ResourceSets:"
    )


async def test_resource_not_exported(snippetcompiler, caplog, environment) -> None:
    """
    test that a warning is logged if a resource that is not exported is in a resource_set
    """
    snippetcompiler.setup_for_snippet(
        """
std::ResourceSet(name="resource_set_1", resources=[std::Resource()])
implement std::Resource using std::none
""",
        autostd=True,
    )
    caplog.clear()
    caplog.set_level(logging.WARNING)
    await snippetcompiler.do_export_and_deploy()
    cwd = snippetcompiler.project_dir

    msg: str = (
        f"resource std::Resource (instantiated at {cwd}/main.cf:2) is part of ResourceSet resource_set_1 "
        f"but will not be exported."
    )

    log_sequence = LogSequence(caplog)
    log_sequence.contains("inmanta.export", logging.WARNING, msg)


async def test_empty_resource_set_removal(snippetcompiler, environment) -> None:
    """
    When a partial compile is ran, the exporter should trigger a deletion of each ResourceSet, defined in the partial model,
    that doesn't have any resources associated
    """

    async def export_model(
        model: str,
        partial_compile: bool,
        resource_sets_to_remove: Optional[list[str]] = None,
    ) -> None:
        snippetcompiler.setup_for_snippet(model, autostd=True)
        await snippetcompiler.do_export_and_deploy(
            partial_compile=partial_compile,
            resource_sets_to_remove=resource_sets_to_remove,
        )

    # Full compile
    await export_model(
        model="""
import test_resources

a = test_resources::Resource(value="A", agent="A", key="the_resource_a")
b = test_resources::Resource(value="A", agent="A", key="the_resource_b")
c = test_resources::Resource(value="A", agent="A", key="the_resource_c")
d = test_resources::Resource(value="A", agent="A", key="the_resource_d")
e = test_resources::Resource(value="A", agent="A", key="the_resource_e")
z = test_resources::Resource(value="A", agent="A", key="the_resource_z")
std::ResourceSet(name="resource_set_1", resources=[a,c])
std::ResourceSet(name="resource_set_2", resources=[b])
std::ResourceSet(name="resource_set_3", resources=[d, e])
        """,
        partial_compile=False,
    )
    await assert_resource_set_assignment(
        environment,
        assignment={
            "the_resource_a": "resource_set_1",
            "the_resource_b": "resource_set_2",
            "the_resource_c": "resource_set_1",
            "the_resource_d": "resource_set_3",
            "the_resource_e": "resource_set_3",
            "the_resource_z": None,
        },
    )

    # Partial compile
    await export_model(
        model="""
    import test_resources


    a = test_resources::Resource(value="A", agent="A", key="the_resource_a")
    c2 = test_resources::Resource(value="A", agent="A", key="the_resource_c2")
    f = test_resources::Resource(value="A", agent="A", key="the_resource_f")
    std::ResourceSet(name="resource_set_1", resources=[a,c2])
    std::ResourceSet(name="resource_set_4", resources=[f])
    std::ResourceSet(name="resource_set_3", resources=[])
            """,
        partial_compile=True,
        resource_sets_to_remove=["resource_set_2"],
    )
    await assert_resource_set_assignment(
        environment,
        assignment={
            "the_resource_a": "resource_set_1",
            "the_resource_c2": "resource_set_1",
            "the_resource_f": "resource_set_4",
            "the_resource_z": None,
        },
    )


def test_attribute_value_of_id_has_str_type(snippetcompiler):
    """
    Verify that the id.attribute_value field of resources emitted by the exporter have the type str.
    Even if the type in the model is different.
    """
    snippetcompiler.setup_for_snippet(
        """
        import testmodule_integer_id

        testmodule_integer_id::Test(
            id_attr=123,
            agent="internal",
        )
        """,
        autostd=True,
    )
    version, resource_dct = snippetcompiler.do_export()
    resources = list(resource_dct.values())
    assert len(resources) == 1
    id_attribute_value = resources[0].id.attribute_value
    assert isinstance(id_attribute_value, str)
    assert id_attribute_value == "123"


async def test_export_duplicate(resource_container, snippetcompiler):
    """
    The exported should provide a compilation error when a resource is defined twice in a model
    """
    snippetcompiler.setup_for_snippet(
        """
        import test

        test::Resource(key="test", value="foo")
        test::Resource(key="test", value="bar")
    """,
        autostd=True,
    )

    with pytest.raises(CompilerException) as exc:
        snippetcompiler.do_export()

    assert "exists more than once in the configuration model" in str(exc.value)


async def test_resource_inherits_from_decorator(snippetcompiler, local_module_package_index):
    """
    Verify that a clear error message is shown to the user if a resource inherits
    from the resource decorator instead of the Resource class.
    """
    requirement = InmantaModuleRequirement.parse("invalid_resource_def")
    snippetcompiler.setup_for_snippet(
        "import invalid_resource_def",
        install_project=True,
        index_url=local_module_package_index,
        python_requires=[requirement.get_python_package_requirement()],
        autostd=False,
    )

    with pytest.raises(RuntimeException) as exc:
        snippetcompiler.do_export()

    assert (
        "Resource inmanta_plugins.invalid_resource_def.Test is inheriting from the inmanta.resources.resource decorator."
        " Did you intend to inherit from inmanta.resources.Resource instead?"
    ) in str(exc.value)



"""
Copyright 2014 Omer Gertel
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This code was originally developed by Omer Gertel, as a python port of the core portion of a
[Java Metrics library by Coda Hale](http://metrics.dropwizard.io/)

It was vendored into the inmanta source tree as the original was no longer maintained.
"""

import heapq
import math
import random
import time

from .. import Clock
from .snapshot import Snapshot

DEFAULT_SIZE = 1028
DEFAULT_ALPHA = 0.015


# TODO: do I ABC this, may cpst us a few % performance??
class Sample:
    def clear(self) -> None:
        raise NotImplementedError()

    def update(self, value: float) -> None:
        raise NotImplementedError()

    def get_snapshot(self) -> Snapshot:
        raise NotImplementedError()


class ExpDecayingSample(Sample):
    """
    An exponentially-decaying random sample of longs. Uses Cormode et al's
    forward-decaying priority reservoir sampling method to produce a
    statistically representative sample, exponentially biased towards newer
    entries.

    @see: <a href="http://www.research.att.com/people/Cormode_Graham/library/publications/CormodeShkapenyukSrivastavaXu09.pdf">
          Cormode et al. Forward Decay: A Practical Time Decay Model for
          Streaming Systems. ICDE '09: Proceedings of the 2009 IEEE
          International Conference on Data Engineering (2009)</a>
    """

    RESCALE_THREASHOLD = 3600.0  # 1 hour

    def __init__(self, size: int = DEFAULT_SIZE, alpha: float = DEFAULT_ALPHA, clock: Clock = time) -> None:
        """
        Creates a new L{ExponentiallyDecayingSample}.

        :type size: C{int}
        :param size: the number of samples to keep in the sampling reservoir
        :type alpha: C{float}
        :param alpha: the exponential decay factor; the higher this is, the more
                      biased the sample will be towards newer values
        :type clock: C{function}
        :param clock: the function used to return the current time, default to
                      seconds since the epoch; to be used with other time
                      units, or with the twisted clock for our testing purposes
        """
        super(ExpDecayingSample, self).__init__()
        self.clock = clock
        self.size = size
        self.alpha = alpha
        self.clear()

    def clear(self) -> None:
        self.values: dict[float, float] = {}
        self.priorities: list[float] = []
        self.counter = 0
        self.start_time = self.clock.time()
        self.next_time = self.clock.time() + ExpDecayingSample.RESCALE_THREASHOLD

    def get_size(self) -> int:
        return self.counter if self.counter < self.size else self.size

    def update(self, value: float) -> None:
        """
        Adds a value to the sample.

        :type value: C{int} or C{float}
        :param value: the value to be added
        """
        if self.size == 0:
            return
        self._rescale_if_necessary()
        priority = self._weight(self.clock.time() - self.start_time) / random.random()
        new_counter = self.counter + 1
        self.counter = new_counter

        if new_counter <= self.size:
            self.values[priority] = value
            heapq.heappush(self.priorities, priority)
        else:
            first = heapq.heappop(self.priorities)
            if first < priority:
                if priority not in self.values:
                    self.values[priority] = value
                    heapq.heappush(self.priorities, priority)
                    while first not in self.values:
                        first = heapq.heappop(self.priorities)
                    del self.values[first]
            else:
                heapq.heappush(self.priorities, first)

    def _rescale_if_necessary(self) -> None:
        if self.clock.time() >= self.next_time:
            self._rescale()

    def _rescale(self) -> None:
        self.next_time = self.clock.time() + ExpDecayingSample.RESCALE_THREASHOLD
        old_start_time = self.start_time
        self.start_time = self.clock.time()
        new_values = {}
        new_priorities: list[float] = []
        for key, val in self.values.items():
            priority = key * math.exp(-self.alpha * (self.start_time - old_start_time))
            new_values[priority] = val
            heapq.heappush(new_priorities, priority)
        self.values = new_values
        self.priorities = new_priorities
        self.counter = len(self.values)

    def _weight(self, value: float) -> float:
        return math.exp(self.alpha * value)

    def get_snapshot(self) -> Snapshot:
        return Snapshot(self.values.values())


class SlidingTimeWindowSample(Sample):
    """
    A sample of measurements made in a sliding time window.
    """

    DEFAULT_WINDOW = 300

    def __init__(self, window: int = DEFAULT_WINDOW, clock: Clock = time) -> None:
        """Creates a SlidingTimeWindowSample.

        :param window: the length of the time window in seconds
        :param clock: clock.time() is called to get the current time as seconds
                      since the epoch.
        """
        self.window = window
        self.clock = clock
        self.clear()

    def clear(self) -> None:
        self.values: list[tuple[float, float]] = []

    def _trim(self) -> None:
        deadline = self.clock.time() - self.window
        while self.values and self.values[0][0] < deadline:
            heapq.heappop(self.values)

    def update(self, value: float) -> None:
        heapq.heappush(self.values, (self.clock.time(), value))

    def get_snapshot(self) -> Snapshot:
        self._trim()
        return Snapshot(x[1] for x in self.values)



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

# pylint: disable-msg=R0923,W0613

import logging
import warnings
from collections.abc import Iterator, Sequence
from typing import Optional

from inmanta.ast import (
    Anchor,
    AttributeAnchor,
    CompilerException,
    CompilerRuntimeWarning,
    DuplicateException,
    HyphenException,
    Import,
    IndexException,
    LocatableString,
    Namespace,
    NotFoundException,
    Range,
    RuntimeException,
    TypeAnchor,
    TypeNotFoundException,
    TypeReferenceAnchor,
    TypingException,
)
from inmanta.ast.attribute import Attribute, RelationAttribute
from inmanta.ast.blocks import BasicBlock
from inmanta.ast.entity import Entity, Implement, Implementation
from inmanta.ast.statements import BiStatement, ExpressionStatement, Literal, Statement, TypeDefinitionStatement
from inmanta.ast.type import TYPES, ConstraintType, NullableType, OrReferenceType, Type, TypedList
from inmanta.execute.runtime import ExecutionUnit, QueueScheduler, Resolver, ResultVariable
from inmanta.plugins import Plugin

from . import DefinitionStatement

LOGGER = logging.getLogger(__name__)


class TypeDeclaration(Statement):
    """
    Declaration of a type. A type declaration consists of a base type string and can be
    multi ('basetype[]'), nullable ('basetype?') or both ('basetype[]?').
    """

    def __init__(
        self,
        basetype: LocatableString,
        multi: bool = False,
        nullable: bool = False,
    ) -> None:
        Statement.__init__(self)
        self.basetype: LocatableString = basetype
        self.multi: bool = multi
        self.nullable: bool = nullable

    def get_basetype(self, namespace: Namespace) -> Type:
        """
        Returns the base type for this declaration as a Type.
        """
        return namespace.get_type(self.basetype)

    def get_type(self, namespace: Namespace) -> Type:
        """
        Returns the type for this declaration as a Type.
        """
        tp: Type = self.get_basetype(namespace)
        if self.multi:
            tp = TypedList(tp)
        if self.nullable:
            tp = NullableType(tp)
        return tp

    def __str__(self) -> str:
        return f"{self.basetype}{'[]' if self.multi else ''}{'?' if self.nullable else ''}"


class DefineAttribute(Statement):
    def __init__(
        self,
        attr_type: TypeDeclaration,
        name: LocatableString,
        default_value: Optional[ExpressionStatement] = None,
        remove_default: bool = True,
    ) -> None:
        """
        if default_value is None, this is an explicit removal of a default value
        """
        super().__init__()
        if "-" in name.value:
            raise HyphenException(name)
        self.type = attr_type
        self.name = name
        self.default = default_value
        self.remove_default = remove_default

    def __str__(self) -> str:
        return f"{self.type} {self.name} = {str(self.default) if self.default else ''}"


class DefineEntity(TypeDefinitionStatement):
    """
    Define a new entity in the configuration
    """

    type: Entity

    def __init__(
        self,
        namespace: Namespace,
        lname: LocatableString,
        comment: Optional[LocatableString],
        parents: list[LocatableString],
        attributes: list[DefineAttribute],
    ) -> None:
        name = str(lname)
        TypeDefinitionStatement.__init__(self, namespace, name)
        if "-" in name:
            raise HyphenException(lname)

        self.name = name
        self.attributes = attributes
        if comment is not None:
            self.comment = str(comment)

        self.parents = parents
        self.anchors = [TypeReferenceAnchor(self.namespace, x) for x in self.parents]

        if len(self.parents) == 0 and not (self.name == "Entity" and self.namespace.name == "std"):
            dummy_location: Range = Range("__internal__", 1, 1, 1, 1)
            self.parents.append(LocatableString("std::Entity", dummy_location, -1, namespace))

        self.type = Entity(self.name, namespace, self.comment)
        self.type.location = lname.location

    def add_attribute(
        self, attr_type: LocatableString, name: LocatableString, default_value: Optional[ExpressionStatement] = None
    ) -> None:
        """
        Add an attribute to this entity
        """
        self.attributes.append(DefineAttribute(TypeDeclaration(attr_type), name, default_value))

    def __repr__(self) -> str:
        """
        A textual representation of this entity
        """
        return "Entity(%s)" % self.name

    def get_full_parent_names(self) -> list[str]:
        def resolve_parent(parent: LocatableString) -> str:
            ptype = self.namespace.get_type(parent)
            assert isinstance(ptype, Entity), f"Parents of entities should be entities, but {parent} is a {type(ptype)}"
            return ptype.get_full_name()

        try:
            return [resolve_parent(parent) for parent in self.parents]
        except TypeNotFoundException as e:
            e.set_statement(self)
            raise e

    def evaluate(self) -> None:
        """
        Evaluate this statement.
        """
        try:
            entity_type = self.type
            entity_type.comment = self.comment
            is_dataclass = False

            add_attributes: dict[str, Attribute] = {}
            attribute: DefineAttribute
            for attribute in self.attributes:
                attr_type: Type = attribute.type.get_type(self.namespace)
                if not isinstance(attr_type, (Type, type)):
                    raise TypingException(self, "Attributes can only be a type. Entities need to be defined as relations.")

                name = str(attribute.name)
                attr_obj = Attribute(
                    entity_type,
                    OrReferenceType(attribute.type.get_basetype(self.namespace)),
                    name,
                    attribute.get_location(),
                    attribute.type.multi,
                    attribute.type.nullable,
                )
                self.anchors.append(TypeReferenceAnchor(self.namespace, attribute.type.basetype))

                if name in add_attributes:
                    raise DuplicateException(attr_obj, add_attributes[name], "Same attribute defined twice in one entity")

                add_attributes[name] = attr_obj

                if attribute.default is not None or attribute.remove_default:
                    entity_type.add_default_value(name, attribute)

            if len({str(p) for p in self.parents}) != len(self.parents):
                raise TypingException(self, "same parent defined twice")
            for parent in self.parents:
                parent_type = self.namespace.get_type(parent)
                if parent_type is self.type:
                    raise TypingException(self, "Entity can not be its own parent (%s) " % parent)
                if not isinstance(parent_type, Entity):
                    raise TypingException(
                        self,
                        "Parents of an entity need to be entities. "
                        "Default constructors are not supported. %s is not an entity" % parent,
                    )

                entity_type.parent_entities.append(parent_type)
                parent_type.child_entities.append(entity_type)

            for parent_type in entity_type.get_all_parent_entities():
                for attr_name, other_attr in parent_type.attributes.items():
                    if attr_name not in add_attributes:
                        add_attributes[attr_name] = other_attr
                    else:
                        # allow compatible attributes
                        my_attr = add_attributes[attr_name]

                        if my_attr.type == other_attr.type:
                            add_attributes[attr_name] = other_attr
                        else:
                            raise DuplicateException(my_attr, other_attr, "Incompatible attributes")
                is_dataclass |= parent_type.get_full_name() == "std::Dataclass"

            if is_dataclass:
                entity_type.pair_dataclass_stage1()
            # verify all attribute compatibility
        except TypeNotFoundException as e:
            e.set_statement(self)
            raise e


class DefineImplementation(TypeDefinitionStatement):
    """
    Define a new implementation that has a name and contains statements

    :param name: The name of the implementation
    """

    type: Implementation

    def __init__(
        self,
        namespace: Namespace,
        name: LocatableString,
        target_type: LocatableString,
        statements: BasicBlock,
        comment: LocatableString,
    ):
        TypeDefinitionStatement.__init__(self, namespace, str(name))
        self.name = str(name)
        if "-" in self.name:
            raise HyphenException(name)

        self.block = statements
        self.entity = target_type

        if comment is not None:
            self.comment = str(comment)

        self.location = name.get_location()

        self.type = Implementation(str(self.name), self.block, self.namespace, str(target_type), self.comment)
        self.type.location = name.get_location()

    def __repr__(self) -> str:
        """
        The representation of this implementation
        """
        return "Implementation(%s)" % self.name

    def evaluate(self) -> None:
        """
        Evaluate this statement in the given scope
        """
        try:
            cls = self.namespace.get_type(self.entity)
            self.anchors = [TypeAnchor(self.entity, cls)]
            if not isinstance(cls, Entity):
                raise TypingException(self, f"Implementation can only be define for an Entity, but {self.entity} is a {cls}")
            self.type.set_type(cls)
            self.copy_location(self.type)
        except TypeNotFoundException as e:
            e.set_statement(self)
            raise e

    def get_anchors(self) -> list[Anchor]:
        """
        This method overrides the default get_anchors() to accommodate the two-stage normalization process.
        DefineImplementations register anchors for their blocks. However, these anchors only come into existence
        after the type normalization phase.
        This implementation ensures that anchors are correctly gathered from both the type statements and
        the block itself.
        """
        return [*self.type.statements.get_anchors(), *self.anchors]

    def nested_blocks(self) -> Iterator["BasicBlock"]:
        """
        Returns an iterator over blocks contained within this statement.
        """
        yield self.block


class DefineImplement(DefinitionStatement):
    """
    Define a new implementation for a given entity

    :param entity: The name of the entity that is implemented
    :param implementations: A list of implementations
    :param select: A clause that determines when this implementation is "active"
    :param inherit: True iff the entity should inherit all implementations from its parents
    """

    comment: Optional[str] = None

    def __init__(
        self,
        entity_name: LocatableString,
        implementations: list[LocatableString],
        select: ExpressionStatement,
        inherit: bool = False,
        comment: Optional[LocatableString] = None,
    ) -> None:
        DefinitionStatement.__init__(self)
        self.entity = entity_name
        self.entity_location = entity_name.get_location()
        self.implementations = implementations
        self.location = entity_name.get_location()
        if inherit and (not isinstance(select, Literal) or select.value is not True):
            raise RuntimeException(self, "Conditional implementation with parents not allowed")
        self.select = select
        self.inherit: bool = inherit
        if comment is not None:
            self.comment = str(comment)

    def __repr__(self) -> str:
        """
        Returns a representation of this class
        """
        return "Implement(%s)" % (self.entity)

    def get_anchors(self) -> list[Anchor]:
        """
        This method overrides the default get_anchors() to accommodate the two-stage normalization process.
        DefineImplement should register anchors for an ExpressionStatement. This one is not yet normalized during
        evaluation and so its anchors come into existence only after the type normalization phase.
        This implementation ensures that anchors are also correctly gathered from the type statement.
        """
        return [*self.anchors, *self.select.get_anchors()]

    def evaluate(self) -> None:
        """
        Evaluate this statement.
        """
        try:
            entity_type = self.namespace.get_type(self.entity)
            if not isinstance(entity_type, Entity):
                raise TypingException(
                    self, f"Implementation can only be define for an Entity, but {self.entity} is a {entity_type}"
                )
            self.anchors.append(TypeAnchor(self.entity, entity_type))

            # If one implements statement has parent declared, set to true
            entity_type.implements_inherits |= self.inherit

            implement = Implement()
            implement.comment = self.comment
            implement.constraint = self.select
            implement.location = self.entity_location

            i = 0
            for _impl in self.implementations:
                i += 1

                # check if the implementation has the correct type
                impl_obj = self.namespace.get_type(_impl)
                assert isinstance(impl_obj, Implementation), "%s is not an implementation" % (_impl)

                if impl_obj.entity is not None and not (
                    entity_type is impl_obj.entity or entity_type.is_parent(impl_obj.entity)
                ):
                    raise TypingException(
                        self,
                        "Type mismatch: cannot use %s as implementation for "
                        " %s because its implementing type is %s" % (impl_obj.name, entity_type, impl_obj.entity),
                    )

                # add it
                self.anchors.append(TypeAnchor(_impl, impl_obj))
                implement.implementations.append(impl_obj)

            entity_type.add_implement(implement)
        except TypeNotFoundException as e:
            e.set_statement(self)
            raise e


class DefineTypeConstraint(TypeDefinitionStatement):
    """
    Define a new data type in the configuration. This type is a constrained
    version of a the built-in datatypes

    :param name: The name of the new  type
    :param basetype: The name of the type that is "refined"
    """

    __expression: ExpressionStatement
    type: ConstraintType

    def __init__(
        self, namespace: Namespace, name: LocatableString, basetype: LocatableString, expression: ExpressionStatement
    ) -> None:
        TypeDefinitionStatement.__init__(self, namespace, str(name))
        self.set_location(name.get_location())
        self.basetype = basetype
        self.set_expression(expression)
        self.type = ConstraintType(self.namespace, str(name))
        self.type.location = name.get_location()
        if self.name in TYPES:
            warnings.warn(CompilerRuntimeWarning(self, "Trying to override a built-in type: %s" % self.name))
        if "-" in self.name:
            raise HyphenException(name)

    def get_expression(self) -> ExpressionStatement:
        """
        Get the expression that constrains the basetype
        """
        return self.__expression

    def set_expression(self, expression: ExpressionStatement) -> None:
        """
        Set the expression that constrains the basetype. This expression
        should reference the value that will be assign to a variable of this
        type. This variable has the same name as the type.
        """
        contains_var = False

        for var in expression.requires():
            if var == self.name or var == "self":
                contains_var = True

        if not contains_var:
            raise TypingException(self, "typedef expressions should reference the self variable")

        self.__expression = expression

    expression: ExpressionStatement = property(get_expression, set_expression)

    def __repr__(self) -> str:
        """
        A representation of this definition
        """
        return "Type(%s)" % self.name

    def evaluate(self) -> None:
        """
        Evaluate this statement.
        """
        basetype = self.namespace.get_type(self.basetype)
        assert not basetype.supports_references()
        self.anchors.append(TypeAnchor(self.basetype, basetype))
        constraint_type = self.type

        constraint_type.comment = self.comment
        constraint_type.basetype = basetype
        constraint_type.constraint = self.expression

    def get_anchors(self) -> list[Anchor]:
        """
        This method overrides the default get_anchors() to accommodate the two-stage normalization process.
        DefineTypeConstraint registers anchors for its condition expression. However, these anchors only come into existence
        after the type normalization phase.
        This implementation ensures that anchors are correctly gathered from both the condition expression and
        the statement itself.
        """
        return [*self.anchors, *self.expression.get_anchors()]


Relationside = tuple[LocatableString, Optional[LocatableString], Optional[tuple[int, Optional[int]]]]


class DefineRelation(BiStatement):
    """
    Define a relation
    """

    annotation_expression: list[tuple[ResultVariable, ExpressionStatement]]

    def __init__(self, left: Relationside, right: Relationside, annotations: list[ExpressionStatement] = []) -> None:
        DefinitionStatement.__init__(self)
        if "-" in str(right[1]):
            raise HyphenException(right[1])

        if "-" in str(left[1]):
            raise HyphenException(left[1])
        # for later evaluation
        self.annotation_expression = [(ResultVariable(), exp) for exp in annotations]
        # for access to results
        self.annotations = [exp[0] for exp in self.annotation_expression]

        self.left: Relationside = left
        self.right: Relationside = right

        self.comment = None

    def __repr__(self) -> str:
        """
        The represenation of this relation
        """
        return f"Relation({self.left[0]}, {self.right[0]})"

    def evaluate(self) -> None:
        """
        Add this relation to the participating ends
        """
        try:
            left = self.namespace.get_type(self.left[0])
        except TypeNotFoundException as e:
            e.set_location(self.location)
            raise e

        assert isinstance(left, Entity), "%s is not an entity" % left

        # Duplicate checking is in entity.normalize
        # Because here we don't know if all entities have been defined

        try:
            right = self.namespace.get_type(self.right[0])
        except TypeNotFoundException as e:
            e.set_location(self.location)
            raise e

        assert isinstance(right, Entity), "%s is not an entity" % right
        # Duplicate checking is in entity.normalize
        # Because here we don't know if all entities have been defined

        left_end: Optional[RelationAttribute]
        if self.left[1] is not None:
            left_end = RelationAttribute(right, left, str(self.left[1]), self.left[1].get_location())
            left_end.target_annotations = self.annotations
            left_end.set_multiplicity(self.left[2])
            left_end.comment = self.comment
        else:
            left_end = None

        right_end: Optional[RelationAttribute]
        if self.right[1] is not None:
            if right == left and str(self.left[1]) == str(self.right[1]):
                # relation is its own inverse
                right_end = left_end
            else:
                right_end = RelationAttribute(left, right, str(self.right[1]), self.right[1].get_location())
                right_end.source_annotations = self.annotations
                right_end.set_multiplicity(self.right[2])
                right_end.comment = self.comment
        else:
            right_end = None

        if left_end is not None and right_end is not None:
            left_end.end = right_end
            right_end.end = left_end

    def emit(self, resolver: Resolver, queue: QueueScheduler) -> None:
        for rv, exp in self.annotation_expression:
            reqs = exp.requires_emit(resolver, queue)
            ExecutionUnit(queue, resolver, rv, reqs, exp)

    def normalize(self) -> None:
        for _, exp in self.annotation_expression:
            exp.normalize()
            self.anchors.extend(exp.get_anchors())

        self.anchors.append(TypeReferenceAnchor(self.left[0].namespace, self.left[0]))
        self.anchors.append(TypeReferenceAnchor(self.right[0].namespace, self.right[0]))


class DefineIndex(DefinitionStatement):
    """
    This defines an index over attributes in an entity
    """

    def __init__(self, entity_type: LocatableString, attributes: list[LocatableString]):
        DefinitionStatement.__init__(self)
        self.type = entity_type
        self.attributes: Sequence[LocatableString] = attributes

    def types(self, recursive: bool = False) -> list[tuple[str, LocatableString]]:
        """
        @see Statement#types
        """
        return [("type", self.type)]

    def __repr__(self) -> str:
        return "index {}({})".format(self.type, ", ".join([str(a) for a in self.attributes]))

    def evaluate(self) -> None:
        """
        Add the index to the entity
        """
        entity_type = self.namespace.get_type(self.type)
        assert isinstance(entity_type, Entity), "%s is not an entity" % entity_type
        self.anchors.append(TypeAnchor(self.type, entity_type))

        allattributes = entity_type.get_all_attribute_names()
        for attribute in self.attributes:
            str_attribute = str(attribute)
            if str_attribute not in allattributes:
                raise NotFoundException(
                    self,
                    str_attribute,
                    f"Attribute '{str_attribute}' referenced in index is not defined in entity {entity_type}",
                )
            else:
                rattribute = entity_type.get_attribute(str_attribute)
                self.anchors.append(AttributeAnchor(attribute.get_location(), rattribute))
                assert rattribute is not None  # Make mypy happy
                if rattribute.is_optional() and isinstance(rattribute, RelationAttribute):
                    raise IndexException(
                        self,
                        f"Index can not contain optional relations, Relation ' {attribute}.{entity_type}' is optional",
                    )
                if rattribute.is_multi():
                    raise IndexException(
                        self, f"Index can not contain list attributes, Attribute ' {attribute}.{entity_type}' is a list"
                    )

        entity_type.add_index([str(a) for a in self.attributes], self)


class PluginStatement(TypeDefinitionStatement):
    """
    This statement defines a plugin function
    """

    def __init__(self, namespace: Namespace, name: str, function_class: type[Plugin]) -> None:
        TypeDefinitionStatement.__init__(self, namespace, name)
        self._name = name
        self._function_class = function_class
        self.type = self._function_class(namespace)

    def __repr__(self) -> str:
        """
        The representation of this function
        """
        return "Function(%s)" % self._name

    def evaluate(self) -> None:
        """
        Evaluate this plugin
        """


class DefineImport(TypeDefinitionStatement, Import):
    def __init__(self, name: LocatableString, toname: LocatableString) -> None:
        DefinitionStatement.__init__(self)
        self.name = str(name)
        if "-" in self.name:
            raise CompilerException(
                "%s is not a valid module name: hyphens are not allowed, please use underscores instead." % (self.name)
            )
        self.toname = str(toname)
        if "-" in self.toname:
            raise HyphenException(toname)

    def register_types(self) -> None:
        self.target = self.namespace.get_ns_from_string(self.name)
        if self.target is None:
            raise TypeNotFoundException(self.name, self.namespace)
        self.namespace.import_ns(self.toname, self)

    def evaluate(self) -> None:
        """
        Evaluate this plugin
        """

    def __str__(self) -> str:
        if self.toname == self.name:
            return f"import {self.name}"
        else:
            return f"import {self.name} as {self.toname}"



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import datetime
import logging
import typing
import uuid
from collections import abc, defaultdict
from collections.abc import Sequence
from typing import Any, Callable, Optional, Union, cast

from asyncpg.connection import Connection
from pydantic import ValidationError
from tornado.httputil import url_concat

from inmanta import const, data, util
from inmanta.const import STATE_UPDATE, TERMINAL_STATES, TRANSIENT_STATES, VALID_STATES_ON_STATE_UPDATE, Change, ResourceState
from inmanta.data import APILIMIT, InvalidSort, model
from inmanta.data.dataview import (
    DiscoveredResourceView,
    ResourceHistoryView,
    ResourceLogsView,
    ResourcesInVersionView,
    ResourceView,
)
from inmanta.data.model import (
    DiscoveredResourceABC,
    LatestReleasedResource,
    LinkedDiscoveredResource,
    ReleasedResourceDetails,
    Resource,
    ResourceAction,
    ResourceHistory,
    ResourceLog,
    VersionedResource,
    VersionedResourceDetails,
)
from inmanta.db.util import ConnectionMaybeInTransaction, ConnectionNotInTransaction
from inmanta.protocol import handle, methods, methods_v2
from inmanta.protocol.common import ReturnValue
from inmanta.protocol.exceptions import BadRequest, Forbidden, NotFound, ServerError
from inmanta.protocol.return_value_meta import ReturnValueWithMeta
from inmanta.resources import Id
from inmanta.server import SLICE_AGENT_MANAGER, SLICE_DATABASE, SLICE_ENVIRONMENT, SLICE_RESOURCE, SLICE_TRANSPORT, agentmanager
from inmanta.server import config as opt
from inmanta.server import extensions, protocol
from inmanta.server.services.environmentlistener import EnvironmentAction, EnvironmentListener
from inmanta.server.validate_filter import InvalidFilter
from inmanta.types import Apireturn, JsonType, PrimitiveTypes, ResourceIdStr, ResourceType, ResourceVersionIdStr
from inmanta.util import parse_timestamp

resource_discovery = extensions.BoolFeature(
    slice=SLICE_RESOURCE,
    name="resource_discovery",
    description="Enable resource discovery. This feature controls the APIs it does not affect the use of discovery resources.",
)


LOGGER = logging.getLogger(__name__)


def error_and_log(message: str, **context: Any) -> None:
    """
    :param message: message to return both to logger and to remote caller
    :param context: additional context to attach to log
    """
    ctx = ",".join([f"{k}: {v}" for k, v in context.items()])
    LOGGER.error("%s %s", message, ctx)
    raise BadRequest(message)


class ResourceActionLogLine(logging.LogRecord):
    """A special log record that is used to report log lines that come from the agent"""

    def __init__(self, logger_name: str, level: int, msg: str, created: datetime.datetime) -> None:
        super().__init__(
            name=logger_name,
            level=level,
            pathname="(unknown file)",
            lineno=0,
            msg=msg,
            args=[],
            exc_info=None,
            func=None,
            sinfo=None,
        )

        self.created = created.timestamp()
        self.msecs = (self.created - int(self.created)) * 1000
        self.relativeCreated = (self.created - logging._startTime) * 1000


class ResourceService(protocol.ServerSlice, EnvironmentListener):
    """Resource Manager service"""

    agentmanager_service: "agentmanager.AgentManager"

    def __init__(self) -> None:
        super().__init__(SLICE_RESOURCE)

        # Dict: environment_id: (model_version, increment, negative_increment, negative_increment_per_agent, run_ahead_lock)
        self._increment_cache: dict[
            uuid.UUID,
            Optional[
                tuple[
                    int,
                    abc.Set[ResourceIdStr],
                    abc.Set[ResourceIdStr],
                    abc.Mapping[str, abc.Set[ResourceIdStr]],
                    Optional[asyncio.Event],
                ]
            ],
        ] = {}
        # lock to ensure only one inflight request
        self._increment_cache_locks: dict[uuid.UUID, asyncio.Lock] = defaultdict(lambda: asyncio.Lock())
        self._resource_action_logger = logging.getLogger(const.NAME_RESOURCE_ACTION_LOGGER)

    def get_dependencies(self) -> list[str]:
        return [SLICE_DATABASE, SLICE_AGENT_MANAGER]

    def get_depended_by(self) -> list[str]:
        return [SLICE_TRANSPORT]

    def define_features(self) -> list[extensions.Feature]:
        return [resource_discovery]

    async def prestart(self, server: protocol.Server) -> None:
        await super().prestart(server)
        self.agentmanager_service = cast("agentmanager.AgentManager", server.get_slice(SLICE_AGENT_MANAGER))
        # This is difficult to type without import loop
        # The type is EnvironmentService
        server.get_slice(SLICE_ENVIRONMENT).register_listener_for_multiple_actions(
            self, [EnvironmentAction.deleted, EnvironmentAction.cleared]
        )

    async def start(self) -> None:
        self.schedule(
            data.ResourceAction.purge_logs, opt.server_purge_resource_action_logs_interval.get(), cancel_on_stop=False
        )
        await super().start()

    async def stop(self) -> None:
        await super().stop()

    def clear_env_cache(self, env: data.Environment | model.Environment) -> None:
        LOGGER.log(const.LOG_LEVEL_TRACE, "Clearing cache for %s", env.id)
        self._increment_cache[env.id] = None

    async def environment_action_cleared(self, env: model.Environment) -> None:
        """
        Will be called when the environment is cleared
        :param env: The environment that is cleared
        """
        self.clear_env_cache(env)

    async def environment_action_deleted(self, env: model.Environment) -> None:
        """
        Will be called when the environment is deleted
        :param env: The environment that is deleted
        """
        self.clear_env_cache(env)

    def get_resource_action_logger(self, environment: uuid.UUID) -> logging.Logger:
        """Get the resource action logger for the given environment.
        :param environment: The environment to get a logger for
        :return: The logger for the given environment.
        """
        return self._resource_action_logger.getChild(str(environment))

    def log_resource_action(
        self, env: uuid.UUID, resource_ids: Sequence[str], log_level: int, ts: datetime.datetime, message: str
    ) -> None:
        """Write the given log to the correct resource action logger"""
        logger = self.get_resource_action_logger(env)
        if len(resource_ids) == 0:
            message = "no resources: " + message
        elif len(resource_ids) > 1:
            message = "multiple resources: " + message
        else:
            message = resource_ids[0] + ": " + message
        log_record = ResourceActionLogLine(logger.name, log_level, message, ts)
        logger.handle(log_record)

    @handle(methods.get_resource, resource_id="id", env="tid")
    async def get_resource(
        self,
        env: data.Environment,
        resource_id: ResourceVersionIdStr,
        logs: bool,
        log_action: const.ResourceAction,
        log_limit: int,
        connection: Optional[Connection] = None,
    ) -> Apireturn:
        # Validate resource version id
        try:
            Id.parse_resource_version_id(resource_id)
        except ValueError:
            return 400, {"message": f"{resource_id} is not a valid resource version id"}

        async with data.ResourceAction.get_connection(connection) as con:
            resv = await data.Resource.get(env.id, resource_id, con)
            if resv is None:
                return 404, {"message": "The resource with the given id does not exist in the given environment"}

            actions: list[data.ResourceAction] = []
            if bool(logs):
                action_name = None
                if log_action is not None:
                    action_name = log_action.name

                actions = await data.ResourceAction.get_log(
                    environment=env.id, resource_version_id=resource_id, action=action_name, limit=log_limit, connection=con
                )

            return 200, {"resource": resv, "logs": actions}

    # This endpoint doesn't have a method associated yet.
    # Intended for use by other slices
    async def get_resources_in_latest_version(
        self,
        environment: data.Environment,
        resource_type: Optional[ResourceType] = None,
        attributes: dict[PrimitiveTypes, PrimitiveTypes] = {},
        connection: Optional[Connection] = None,
    ) -> list[Resource]:
        result = await data.Resource.get_resources_in_latest_version(
            environment.id, resource_type, attributes, connection=connection
        )
        return [r.to_dto() for r in result]

    async def mark_deployed(
        self,
        env: data.Environment,
        resources_id: abc.Set[ResourceIdStr],
        timestamp: datetime.datetime,
        version: int,
        filter: Callable[[ResourceIdStr], bool] = lambda x: True,
        connection: ConnectionMaybeInTransaction = ConnectionNotInTransaction(),
        only_update_from_states: Optional[Sequence[const.ResourceState]] = None,
    ) -> None:
        """
        Set the status of the provided resources as deployed
        :param env: Environment to consider.
        :param resources_id: Set of resources to mark as deployed.
        :param timestamp: Timestamp for the log message and the resource action entry.
        :param version: Version of the resources to consider.
        :param filter: Filter function that takes a resource id as an argument and returns True if it should be kept.
        """
        if not resources_id:
            return

        # performance-critical path: avoid parsing cost if we can
        resources_id_filtered = [res_id for res_id in resources_id if filter(res_id)]
        if not resources_id_filtered:
            return

        action_id = uuid.uuid4()

        async with data.Resource.get_connection(connection.connection) as inner_connection:
            async with inner_connection.transaction():
                # validate resources
                if only_update_from_states is not None:
                    resources = await data.Resource.get_resource_ids_with_status(
                        env.id,
                        resources_id_filtered,
                        version,
                        only_update_from_states,
                        # acquire lock on Resource before read and before lock on ResourceAction to prevent conflicts with
                        # cascading deletes
                        lock=data.RowLockMode.FOR_NO_KEY_UPDATE,
                        connection=inner_connection,
                    )
                    if not resources:
                        return None

                resources_version_ids: list[ResourceVersionIdStr] = [
                    ResourceVersionIdStr(f"{res_id},v={version}") for res_id in resources_id_filtered
                ]

                resource_action = data.ResourceAction(
                    environment=env.id,
                    version=version,
                    resource_version_ids=resources_version_ids,
                    action_id=action_id,
                    action=const.ResourceAction.deploy,
                    started=timestamp,
                    messages=[
                        {
                            "level": "INFO",
                            "msg": "Setting deployed due to known good status",
                            "args": [],
                            "timestamp": timestamp.isoformat(timespec="microseconds"),
                        }
                    ],
                    changes={},
                    status=const.ResourceState.deployed,
                    change=const.Change.nochange,
                    finished=timestamp,
                )
                await resource_action.insert(connection=inner_connection)
                self.log_resource_action(
                    env.id,
                    resources_version_ids,
                    const.LogLevel.INFO.to_int,
                    timestamp,
                    "Setting deployed due to known good status",
                )

                await data.Resource.set_deployed_multi(env.id, resources_id_filtered, version, connection=inner_connection)
                # Resource persistent state should not be affected

    async def get_increment(
        self,
        env: data.Environment,
        version: int,
        connection: Optional[Connection] = None,
        run_ahead_lock: Optional[asyncio.Event] = None,
    ) -> tuple[int, abc.Set[ResourceIdStr], abc.Set[ResourceIdStr], abc.Mapping[str, abc.Set[ResourceIdStr]]]:
        """
        Get the increment for a given environment and a given version of the model from the _increment_cache if possible.
        In case of cache miss, the increment calculation is performed behind a lock to make sure it is only done once per
        version, per environment.

        :param env: The environment to consider.
        :param version: The version of the model to consider.
        :param connection: connection to use towards the DB.
            When the connection is in a transaction, we will always invalidate the cache
        :param run_ahead_lock: lock used to keep agents hanging while building up the latest version
        """

        async def _get_cache_entry() -> (
            Optional[tuple[int, abc.Set[ResourceIdStr], abc.Set[ResourceIdStr], abc.Mapping[str, abc.Set[ResourceIdStr]]]]
        ):
            """
            Returns a tuple (increment, negative_increment, negative_increment_per_agent)
            if a cache entry exists for the given environment and version
            or None if no such cache entry exists.
            """
            cache_entry = self._increment_cache.get(env.id, None)
            if cache_entry is None:
                # No cache entry found
                return None
            (version_cache_entry, incr, neg_incr, neg_incr_per_agent, cached_run_ahead_lock) = cache_entry
            if version_cache_entry >= version:
                assert not run_ahead_lock  # We only expect a lock if WE are ahead
                # Cache is ahead or equal
                if cached_run_ahead_lock is not None:
                    await cached_run_ahead_lock.wait()
            elif version_cache_entry != version:
                # Cache entry exists for another version
                # Expire
                return None
            return version_cache_entry, incr, neg_incr, neg_incr_per_agent

        increment: Optional[
            tuple[int, abc.Set[ResourceIdStr], abc.Set[ResourceIdStr], abc.Mapping[str, abc.Set[ResourceIdStr]]]
        ] = await _get_cache_entry()
        if increment is None or (connection is not None and connection.is_in_transaction()):
            lock = self._increment_cache_locks[env.id]
            async with lock:
                increment = await _get_cache_entry()
                if increment is None:
                    positive, negative = await data.ConfigurationModel.get_increment(env.id, version, connection=connection)
                    negative_per_agent: dict[str, set[ResourceIdStr]] = defaultdict(set)
                    for rid in negative:
                        negative_per_agent[Id.parse_id(rid).agent_name].add(rid)
                    increment = (version, positive, negative, negative_per_agent)
                    self._increment_cache[env.id] = (version, positive, negative, negative_per_agent, run_ahead_lock)
        return increment

    @handle(methods.resource_action_update, env="tid")
    async def resource_action_update(
        self,
        env: data.Environment,
        resource_ids: list[ResourceVersionIdStr],
        action_id: uuid.UUID,
        action: const.ResourceAction,
        started: datetime.datetime,
        finished: datetime.datetime,
        status: Optional[Union[const.ResourceState, const.DeprecatedResourceState]],
        messages: list[dict[str, Any]],
        changes: dict[ResourceVersionIdStr, dict[str, object]],
        change: const.Change,
        send_events: bool,
        keep_increment_cache: bool = False,
        only_update_from_states: Optional[set[const.ResourceState]] = None,
        *,
        connection: ConnectionMaybeInTransaction = ConnectionNotInTransaction(),
    ) -> Apireturn:
        def convert_legacy_state(
            status: Optional[Union[const.ResourceState, const.DeprecatedResourceState]],
        ) -> Optional[const.ResourceState]:
            if status is None or isinstance(status, const.ResourceState):
                return status
            if status == const.DeprecatedResourceState.processing_events:
                return const.ResourceState.deploying
            else:
                raise BadRequest(f"Unsupported deprecated resources state {status.value}")

        # TODO: get rid of this?
        status = convert_legacy_state(status)

        # can update resource state
        is_resource_state_update = action in STATE_UPDATE
        # this ra is finishing
        is_resource_action_finished = finished is not None

        if is_resource_state_update:
            # if status update, status is required
            if status is None:
                error_and_log(
                    "Cannot perform state update without a status.",
                    resource_ids=resource_ids,
                    action=action,
                    action_id=action_id,
                )
            # and needs to be valid
            if status not in VALID_STATES_ON_STATE_UPDATE:
                error_and_log(
                    f"Status {status} is not valid on action {action}",
                    resource_ids=resource_ids,
                    action=action,
                    action_id=action_id,
                )
            if status in TRANSIENT_STATES:
                if not is_resource_action_finished:
                    pass
                else:
                    error_and_log(
                        "The finished field must not be set for transient states",
                        status=status,
                        resource_ids=resource_ids,
                        action=action,
                        action_id=action_id,
                    )
            else:
                if is_resource_action_finished:
                    pass
                else:
                    error_and_log(
                        "The finished field must be set for none transient states",
                        status=status,
                        resource_ids=resource_ids,
                        action=action,
                        action_id=action_id,
                    )

        assert all(Id.is_resource_version_id(rvid) for rvid in resource_ids)

        resources: list[data.Resource]
        async with data.Resource.get_connection(connection.connection) as inner_connection:
            async with inner_connection.transaction():
                # validate resources
                resources = await data.Resource.get_resources(
                    env.id,
                    resource_ids,
                    # acquire lock on Resource before read and before lock on ResourceAction to prevent conflicts with
                    # cascading deletes
                    lock=data.RowLockMode.FOR_NO_KEY_UPDATE,
                    connection=inner_connection,
                )
                if len(resources) == 0 or (len(resources) != len(resource_ids)):
                    raise NotFound(
                        message="The resources with the given ids do not exist in the given environment. "
                        f"Only {len(resources)} of {len(resource_ids)} resources found."
                    )

                if only_update_from_states is not None:
                    resources = [resource for resource in resources if resource.status in only_update_from_states]
                    if not resources:
                        return 200, {"message": "no resources with the given state found"}
                    resource_ids = [resource.resource_version_id for resource in resources]

                # validate transitions
                if is_resource_state_update:
                    # no escape from terminal
                    if any(resource.status != status and resource.status in TERMINAL_STATES for resource in resources):
                        LOGGER.error("Attempting to set undeployable resource to deployable state")
                        raise AssertionError("Attempting to set undeployable resource to deployable state")

                version = Id.parse_id(resource_ids[0]).version

                # get instance
                resource_action = await data.ResourceAction.get(action_id=action_id, connection=inner_connection)
                if resource_action is None:
                    # new
                    if started is None:
                        raise ServerError(message="A resource action can only be created with a start datetime.")

                    resource_action = data.ResourceAction(
                        environment=env.id,
                        version=version,
                        resource_version_ids=resource_ids,
                        action_id=action_id,
                        action=action,
                        started=started,
                    )
                    await resource_action.insert(connection=inner_connection)
                else:
                    # existing
                    if resource_action.finished is not None:
                        raise ServerError(
                            message="An resource action can only be updated when it has not been finished yet. This action "
                            f"finished at {resource_action.finished}"
                        )
                for msg in messages:
                    # All other data is stored in the database. The msg was already formatted at the client side.
                    self.log_resource_action(
                        env.id,
                        resource_ids,
                        const.LogLevel(msg["level"]).to_int,
                        parse_timestamp(msg["timestamp"]),
                        msg["msg"],
                    )
                await resource_action.set_and_save(
                    messages=[
                        {
                            **msg,
                            "timestamp": parse_timestamp(msg["timestamp"]).isoformat(timespec="microseconds"),
                        }
                        for msg in messages
                    ],
                    changes=changes,
                    status=status,
                    change=change,
                    finished=finished,
                    connection=inner_connection,
                )

                if is_resource_state_update:
                    # transient resource update
                    if not is_resource_action_finished:
                        for res in resources:
                            await res.update_fields(status=status, connection=inner_connection)
                        if not keep_increment_cache:
                            self.clear_env_cache(env)
                        return 200

                    else:
                        # final resource update
                        if not keep_increment_cache:
                            self.clear_env_cache(env)

                        for res in resources:
                            await res.update_fields(
                                status=status,
                                connection=inner_connection,
                            )
                            # Not very typeable
                            extra_fields: dict[str, Any] = {}

                            if change is not None and change != Change.nochange:
                                extra_fields["last_produced_events"] = finished

                            if status == ResourceState.deployed:
                                extra_fields["last_success"] = resource_action.started
                            if status not in {
                                ResourceState.deploying,
                                ResourceState.undefined,
                                ResourceState.skipped_for_undefined,
                            }:
                                extra_fields["last_non_deploying_status"] = const.NonDeployingResourceState(status)
                                extra_fields["last_deployed_attribute_hash"] = res.attribute_hash
                                extra_fields["last_deploy"] = finished
                                extra_fields["last_deployed_version"] = version

                            await res.update_persistent_state(
                                **extra_fields,
                                connection=inner_connection,
                            )

                            if (
                                "purged" in res.attributes
                                and res.attributes["purged"]
                                and status == const.ResourceState.deployed
                            ):
                                await data.Parameter.delete_all(
                                    environment=env.id, resource_id=res.resource_id, connection=inner_connection
                                )

        if is_resource_state_update and is_resource_action_finished:

            def post_deploy_update() -> None:
                waiting_agents = {
                    (Id.parse_id(prov).get_agent_name(), res.resource_version_id) for res in resources for prov in res.provides
                }

                for agent, resource_id in waiting_agents:
                    aclient = self.agentmanager_service.get_agent_client(env.id, agent)
                    if aclient is not None:
                        if change is None:
                            my_change = const.Change.nochange
                        else:
                            my_change = change

                        self.add_background_task(
                            aclient.resource_event(env.id, agent, resource_id, send_events, status, my_change, changes)
                        )

            connection.call_after_tx(post_deploy_update)

        return 200

    @handle(methods_v2.get_resource_actions, env="tid")
    async def get_resource_actions(
        self,
        env: data.Environment,
        resource_type: Optional[str] = None,
        agent: Optional[str] = None,
        attribute: Optional[str] = None,
        attribute_value: Optional[str] = None,
        log_severity: Optional[str] = None,
        limit: Optional[int] = 0,
        action_id: Optional[uuid.UUID] = None,
        first_timestamp: Optional[datetime.datetime] = None,
        last_timestamp: Optional[datetime.datetime] = None,
        exclude_changes: Optional[list[Change]] = None,
    ) -> ReturnValue[list[ResourceAction]]:
        if exclude_changes is None:
            exclude_changes = []

        if (attribute and not attribute_value) or (not attribute and attribute_value):
            raise BadRequest(
                f"Attribute and attribute_value should both be supplied to use them filtering. "
                f"Received attribute: {attribute}, attribute_value: {attribute_value}"
            )
        if first_timestamp and last_timestamp:
            raise BadRequest(
                f"Only one of the parameters first_timestamp and last_timestamp should be used "
                f"Received first_timestamp: {first_timestamp}, last_timestamp: {last_timestamp}"
            )
        if action_id and not (first_timestamp or last_timestamp):
            raise BadRequest(
                f"The action_id parameter should be used in combination with either the first_timestamp or the last_timestamp "
                f"Received action_id: {action_id}, first_timestamp: {first_timestamp}, last_timestamp: {last_timestamp}"
            )

        if limit is None:
            limit = APILIMIT
        elif limit > APILIMIT:
            raise BadRequest(f"limit parameter can not exceed {APILIMIT}, got {limit}.")

        resource_actions = await data.ResourceAction.query_resource_actions(
            env.id,
            resource_type,
            agent,
            attribute=attribute,
            attribute_value=attribute_value,
            log_severity=log_severity,
            limit=limit,
            action_id=action_id,
            first_timestamp=first_timestamp,
            last_timestamp=last_timestamp,
            exclude_changes=exclude_changes,
        )
        resource_action_dtos = [resource_action.to_dto() for resource_action in resource_actions]
        links = {}

        def _get_query_params(
            resource_type: Optional[str] = None,
            agent: Optional[str] = None,
            attribute: Optional[str] = None,
            attribute_value: Optional[str] = None,
            log_severity: Optional[str] = None,
            limit: Optional[int] = 0,
        ) -> dict[str, str]:
            query_params = {
                "resource_type": resource_type,
                "agent": agent,
                "attribute": attribute,
                "attribute_value": attribute_value,
                "log_severity": log_severity,
                "limit": str(limit) if limit else None,
            }
            return {param_key: param_value for param_key, param_value in query_params.items() if param_value is not None}

        if limit and resource_action_dtos:
            base_url = "/api/v2/resource_actions"
            common_query_params = _get_query_params(resource_type, agent, attribute, attribute_value, log_severity, limit)
            # Next is always earlier with regards to 'started' time
            next_params = common_query_params.copy()
            next_params["last_timestamp"] = util.datetime_iso_format(resource_action_dtos[-1].started)
            next_params["action_id"] = str(resource_action_dtos[-1].action_id)
            links["next"] = url_concat(base_url, next_params)
            previous_params = common_query_params.copy()
            previous_params["first_timestamp"] = util.datetime_iso_format(resource_action_dtos[0].started)
            previous_params["action_id"] = str(resource_action_dtos[0].action_id)
            links["prev"] = url_concat(base_url, previous_params)
        return_value = ReturnValue(response=resource_action_dtos, links=links if links else None)
        return return_value

    @handle(methods_v2.get_resource_events, env="tid", resource_id="rvid")
    async def get_resource_events(
        self, env: data.Environment, resource_id: Id, exclude_change: Optional[const.Change] = None
    ) -> dict[ResourceIdStr, list[ResourceAction]]:
        return {
            k: [ra.to_dto() for ra in v]
            for k, v in (await data.ResourceAction.get_resource_events(env, resource_id, exclude_change)).items()
        }

    @handle(methods_v2.resource_did_dependency_change, env="tid", resource_id="rvid")
    async def resource_did_dependency_change(
        self,
        env: data.Environment,
        resource_id: Id,
    ) -> bool:
        # This resource has been deployed before => determine whether it should be redeployed based on events
        return any(
            True
            for dependency, actions in (
                await data.ResourceAction.get_resource_events(env, resource_id, const.Change.nochange)
            ).items()
            for action in actions
        )

    @handle(methods_v2.resource_list, env="tid")
    async def resource_list(
        self,
        env: data.Environment,
        limit: Optional[int] = None,
        first_id: Optional[ResourceVersionIdStr] = None,
        last_id: Optional[ResourceVersionIdStr] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        filter: Optional[dict[str, list[str]]] = None,
        sort: str = "resource_type.desc",
        deploy_summary: bool = False,
    ) -> ReturnValueWithMeta[Sequence[LatestReleasedResource]]:
        try:
            handler = ResourceView(env, limit, first_id, last_id, start, end, filter, sort, deploy_summary)

            out: ReturnValueWithMeta[Sequence[LatestReleasedResource]] = await handler.execute()
            if deploy_summary:
                out.metadata["deploy_summary"] = await data.Resource.get_resource_deploy_summary(env.id)
            return out

        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

        # TODO: optimize for no orphans

    @handle(methods_v2.resource_details, env="tid")
    async def resource_details(self, env: data.Environment, rid: ResourceIdStr) -> ReleasedResourceDetails:
        details = await data.Resource.get_released_resource_details(env.id, rid)
        if not details:
            raise NotFound("The resource with the given id does not exist, or was not released yet in the given environment.")
        return details

    @handle(methods_v2.resource_history, env="tid")
    async def resource_history(
        self,
        env: data.Environment,
        rid: ResourceIdStr,
        limit: Optional[int] = None,
        first_id: Optional[str] = None,
        last_id: Optional[str] = None,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        sort: str = "date.desc",
    ) -> ReturnValue[Sequence[ResourceHistory]]:
        try:
            handler = ResourceHistoryView(
                environment=env,
                rid=rid,
                limit=limit,
                sort=sort,
                first_id=first_id,
                last_id=last_id,
                start=start,
                end=end,
            )
            out = await handler.execute()
            return out
        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

    @handle(methods_v2.resource_logs, env="tid")
    async def resource_logs(
        self,
        env: data.Environment,
        rid: ResourceIdStr,
        limit: Optional[int] = None,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        filter: Optional[dict[str, list[str]]] = None,
        sort: str = "timestamp.desc",
    ) -> ReturnValue[Sequence[ResourceLog]]:
        try:
            handler = ResourceLogsView(environment=env, rid=rid, limit=limit, sort=sort, start=start, end=end, filter=filter)
            out = await handler.execute()
            return out
        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

    @handle(methods_v2.get_resources_in_version, env="tid")
    async def get_resources_in_version(
        self,
        env: data.Environment,
        version: int,
        limit: Optional[int] = None,
        first_id: Optional[ResourceVersionIdStr] = None,
        last_id: Optional[ResourceVersionIdStr] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        filter: Optional[dict[str, list[str]]] = None,
        sort: str = "resource_type.desc",
    ) -> ReturnValueWithMeta[Sequence[VersionedResource]]:
        try:
            handler = ResourcesInVersionView(env, version, limit, filter, sort, first_id, last_id, start, end)
            return await handler.execute()
        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

    @handle(methods_v2.versioned_resource_details, env="tid")
    async def versioned_resource_details(
        self, env: data.Environment, version: int, rid: ResourceIdStr
    ) -> VersionedResourceDetails:
        resource = await data.Resource.get_versioned_resource_details(environment=env.id, version=version, resource_id=rid)
        if not resource:
            raise NotFound("The resource with the given id does not exist")
        return resource

    @handle(methods_v2.discovered_resource_create, env="tid")
    async def discovered_resource_create(
        self,
        env: data.Environment,
        discovered_resource_id: ResourceIdStr,
        values: JsonType,
        discovery_resource_id: ResourceIdStr,
    ) -> None:
        try:
            discovered_resource = LinkedDiscoveredResource(
                discovered_resource_id=discovered_resource_id, values=values, discovery_resource_id=discovery_resource_id
            )
        except ValidationError as e:
            # this part was copy/pasted from protocol.common.MethodProperties.validate_arguments.
            error_msg = f"Failed to validate argument\n{str(e)}"
            LOGGER.exception(error_msg)
            raise BadRequest(error_msg, {"validation_errors": e.errors()})

        dao = discovered_resource.to_dao(env.id)
        await dao.insert_with_overwrite()

    @handle(methods_v2.discovered_resource_create_batch, env="tid")
    async def discovered_resources_create_batch(
        self, env: data.Environment, discovered_resources: list[LinkedDiscoveredResource]
    ) -> None:
        dao_list = [res.to_dao(env.id) for res in discovered_resources]
        await data.DiscoveredResource.insert_many_with_overwrite(dao_list)

    @handle(methods_v2.discovered_resources_get, env="tid")
    async def discovered_resources_get(
        self, env: data.Environment, discovered_resource_id: ResourceIdStr
    ) -> DiscoveredResourceABC:
        if not self.feature_manager.enabled(resource_discovery):
            raise Forbidden(message="The resource discovery feature is not enabled.")

        result = await data.DiscoveredResource.get_one(environment=env.id, discovered_resource_id=discovered_resource_id)
        if not result:
            raise NotFound(f"discovered_resource with name {discovered_resource_id} not found in env {env.id}")
        dto = result.to_dto()
        return dto

    @protocol.handle(methods_v2.discovered_resources_get_batch, env="tid")
    async def discovered_resources_get_batch(
        self,
        env: data.Environment,
        limit: Optional[int] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        sort: str = "discovered_resource_id.asc",
        filter: Optional[dict[str, list[str]]] = None,
    ) -> ReturnValue[Sequence[DiscoveredResourceABC]]:
        if not self.feature_manager.enabled(resource_discovery):
            raise Forbidden(message="The resource discovery feature is not enabled.")

        try:
            handler = DiscoveredResourceView(environment=env, limit=limit, sort=sort, start=start, end=end, filter=filter)
            out = await handler.execute()

            return out
        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

    @handle(methods_v2.discovered_resource_delete, env="tid")
    async def discovered_resource_delete(self, env: data.Environment, discovered_resource_id: ResourceIdStr) -> None:
        """
        Deletes a discovered resource based on id.

        :raise NotFound: This exception is raised if the discovered resource is not found in the provided environment
        """
        async with data.Resource.get_connection() as connection:
            result = await data.DiscoveredResource.get_one(
                environment=env.id, discovered_resource_id=discovered_resource_id, connection=connection
            )
            if not result:
                raise NotFound(f"Discovered Resource with id {discovered_resource_id} not found in env {env.id}")
            await result.delete(connection=connection)

    @handle(methods_v2.discovered_resource_delete_batch, env="tid")
    async def discovered_resource_delete_batch(
        self, env: data.Environment, discovered_resource_ids: typing.Sequence[str]
    ) -> None:
        """
        Deletes one or more discovered resources based on id.
        Does not raise an Exception if one or more of the discovered resources are not found.
        """
        async with data.Resource.get_connection() as connection:
            query = f"""
            DELETE FROM {data.DiscoveredResource.table_name()} as dr
            WHERE dr.environment=$1
                AND dr.discovered_resource_id=ANY($2)
            """
            await connection.execute(query, env.id, discovered_resource_ids)



"""
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    """
    * Add foreign key from the notification to the compile table.
    * Remove notifications that have a reference to a compile that was removed.
    """
    schema = """
    -- Add compile_id column to notification table.
    ALTER TABLE public.notification
        ADD COLUMN compile_id UUID;

    -- Populate compile_id column.
    UPDATE public.notification
    SET compile_id=regexp_replace(uri, '^/api/v2/compilereport/', '')::uuid
    WHERE uri ~ '^/api/v2/compilereport/';

    -- Delete notifications that reference a compile that no longer exists in
    -- the database.
    DELETE FROM public.notification AS n
    WHERE n.compile_id IS NOT NULL AND NOT EXISTS(
        SELECT *
        FROM public.compile AS c
        WHERE c.id=n.compile_id
    );

    -- Add back the foreign key constraint now that the update + delete are done
    ALTER TABLE public.notification ADD CONSTRAINT notification_compile_id_fkey
        FOREIGN KEY (compile_id) REFERENCES public.compile(id) ON DELETE CASCADE;

    """
    await connection.execute(schema)



"""
 Copyright 2024 Inmanta

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

 Contact: code@inmanta.com

Async support for @functools.lrucache
From https://github.com/python/cpython/issues/90780
"""

import asyncio
from asyncio import Future
from functools import _lru_cache_wrapper, lru_cache, wraps
from typing import Any, Awaitable, Callable, Generator, ParamSpec, TypeVar, overload

T = TypeVar("T")
_PWrapped = ParamSpec("_PWrapped")


class CachedAwaitable(Awaitable[T]):
    def __init__(self, awaitable: Awaitable[T]) -> None:
        self.awaitable = awaitable
        self.result: Future[T] | None = None

    def __await__(self) -> Generator[Any, None, T]:
        if self.result is None:
            fut = asyncio.get_event_loop().create_future()
            self.result = fut
            try:
                result = yield from self.awaitable.__await__()
                fut.set_result(result)
            except Exception as e:
                fut.set_exception(e)
        if not self.result.done():
            yield from self.result
        return self.result.result()


def reawaitable(func: Callable[_PWrapped, Awaitable[T]]) -> Callable[_PWrapped, CachedAwaitable[T]]:
    @wraps(func)
    def wrapper(*args, **kwargs):
        return CachedAwaitable(func(*args, **kwargs))

    return wrapper


@overload
def async_lru_cache(
    maxsize: int | None = 128, typed: bool = False
) -> "Callable[[Callable[..., Awaitable[T]]], _lru_cache_wrapper[CachedAwaitable[T]]]": ...
@overload
def async_lru_cache(maxsize: Callable[..., Awaitable[T]], typed: bool = False) -> "_lru_cache_wrapper[CachedAwaitable[T]]": ...


def async_lru_cache(
    maxsize: int | None | Callable[..., Awaitable[T]] = 128, typed=False
) -> "Callable[[Callable[..., Awaitable[T]]], _lru_cache_wrapper[CachedAwaitable[T]]] | _lru_cache_wrapper[CachedAwaitable[T]]":
    if callable(maxsize) and isinstance(typed, bool):
        user_function, maxsize = maxsize, 128
        return lru_cache(maxsize, typed)(reawaitable(user_function))

    def decorating_function(user_function):
        return lru_cache(maxsize, typed)(reawaitable(user_function))

    return decorating_function



"""
Copyright 2014 Omer Gertel
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This code was originally developed by Omer Gertel, as a python port of the core portion of a
[Java Metrics library by Coda Hale](http://metrics.dropwizard.io/)

It was vendored into the inmanta source tree as the original was no longer maintained.
"""

import abc
import time
from types import TracebackType
from typing import Optional, Type

from .. import Clock
from ..stats.samples import DEFAULT_ALPHA, DEFAULT_SIZE, Sample
from ..stats.snapshot import Snapshot
from .histogram import Histogram
from .meter import Meter


class TimerSink(abc.ABC):

    @abc.abstractmethod
    def add(self, value: float) -> None:
        pass


class Timer(object):
    """
    A timer metric which aggregates timing durations and provides duration statistics, plus
    throughput statistics via Meter and Histogram.

    """

    def __init__(
        self,
        size: int = DEFAULT_SIZE,
        alpha: float = DEFAULT_ALPHA,
        clock: Clock = time,
        sink: TimerSink | None = None,
        sample: Sample | None = None,
    ) -> None:
        super(Timer, self).__init__()
        self.meter = Meter(clock=clock)
        self.hist = Histogram(size=size, alpha=alpha, clock=clock, sample=sample)
        self.sink = sink

    def get_count(self) -> float:
        "get count from internal histogram"
        return self.hist.get_count()

    def get_sum(self) -> float:
        "get sum from snapshot of internal histogram"
        return self.get_snapshot().get_sum()

    def get_max(self) -> float:
        "get max from snapshot of internal histogram"
        return self.get_snapshot().get_max()

    def get_min(self) -> float:
        "get min from snapshot of internal histogram"
        return self.get_snapshot().get_min()

    def get_mean(self) -> float:
        "get mean from snapshot of internal histogram"
        return self.get_snapshot().get_mean()

    def get_stddev(self) -> float:
        "get stddev from snapshot of internal histogram"
        return self.get_snapshot().get_stddev()

    def get_var(self) -> float:
        "get var from snapshot of internal histogram"
        return self.get_snapshot().get_var()

    def get_snapshot(self) -> Snapshot:
        "get snapshot from internal histogram"
        return self.hist.get_snapshot()

    def get_mean_rate(self) -> float:
        "get mean rate from internal meter"
        return self.meter.get_mean_rate()

    def get_one_minute_rate(self) -> float:
        "get 1 minut rate from internal meter"
        return self.meter.get_one_minute_rate()

    def get_five_minute_rate(self) -> float:
        "get 5 minute rate from internal meter"
        return self.meter.get_five_minute_rate()

    def get_fifteen_minute_rate(self) -> float:
        "get 15 rate from internal meter"
        return self.meter.get_fifteen_minute_rate()

    def _update(self, seconds: float) -> None:
        if seconds >= 0:
            self.hist.add(seconds)
            self.meter.mark()
            if self.sink:
                self.sink.add(seconds)

    def time(self, *args: object, **kwargs: object) -> "TimerContext":
        """
        Parameters will be sent to signal, if fired.
        Returns a timer context instance which can be used from a with-statement.
        Without with-statement you have to call the stop method on the context
        """
        return TimerContext(self, self.meter.clock, *args, **kwargs)

    def clear(self) -> None:
        "clear internal histogram and meter"
        self.hist.clear()
        self.meter.clear()


class TimerContext(object):
    def __init__(self, timer: Timer, clock: Clock, *args: object, **kwargs: object) -> None:
        super(TimerContext, self).__init__()
        self.clock = clock
        self.timer = timer
        self.start_time = self.clock.time()
        self.kwargs = kwargs
        self.args = args

    def stop(self) -> float:
        elapsed: float = self.clock.time() - self.start_time
        self.timer._update(elapsed)
        return elapsed

    def __enter__(self) -> None:
        pass

    def __exit__(self, t: Optional[Type[BaseException]], v: Optional[BaseException], tb: Optional[TracebackType]) -> None:
        self.stop()



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import contextlib
import dataclasses
import datetime
import enum
import itertools
import json
import uuid
from collections import defaultdict
from collections.abc import Mapping, Sequence, Set
from dataclasses import dataclass
from enum import StrEnum
from typing import Optional, Self, cast

import asyncpg

from inmanta import const, resources
from inmanta.types import ResourceIdStr, ResourceType
from inmanta.util.collections import BidirectionalManyMapping


class RequiresProvidesMapping(BidirectionalManyMapping["ResourceIdStr", "ResourceIdStr"]):
    def requires_view(self) -> Mapping["ResourceIdStr", Set["ResourceIdStr"]]:
        """
        Returns a view of the requires relationship of this RequiresProvidesMapping. This view will be updated
        whenever the RequiresProvidesMapping object is updated.
        """
        return self

    def provides_view(self) -> Mapping["ResourceIdStr", Set["ResourceIdStr"]]:
        """
        Returns a view of the provides relationship of this RequiresProvidesMapping. This view will be updated
        whenever the RequiresProvidesMapping object is updated.
        """
        return self.reverse_mapping()


class Compliance(StrEnum):
    """
    Status of a resource's operational status with respect to its latest desired state, to the best of our knowledge.
    COMPLIANT: The operational state complies to latest resource intent as far as we know.
    HAS_UPDATE: The resource intent has been updated since latest deploy attempt (if any),
        meaning we are not yet managing the new intent.
    NON_COMPLIANT: The resource intent has not been updated since latest deploy attempt (if any)
        but we have reason to believe operational state might not comply with latest resource intent,
        based on a deploy attempt / compliance check for that intent.
    UNDEFINED: The resource status is undefined, because it has an unknown attribute. i.e. undefined status <=> undefined
        intent.
    """

    COMPLIANT = enum.auto()
    HAS_UPDATE = enum.auto()
    NON_COMPLIANT = enum.auto()
    UNDEFINED = enum.auto()

    def is_dirty(self) -> bool:
        """
        Return True iff the status indicates the resource is not up-to-date and is ready to be deployed.
        """
        return self in [Compliance.HAS_UPDATE, Compliance.NON_COMPLIANT]


@dataclass(frozen=True)
class ResourceIntent:
    resource_id: "ResourceIdStr"
    attribute_hash: str
    attributes: Mapping[str, object] = dataclasses.field(hash=False)

    id: "resources.Id" = dataclasses.field(init=False, compare=False, hash=False)

    def __post_init__(self) -> None:
        # use object.__setattr__ because this is a frozen dataclass, see dataclasses docs
        object.__setattr__(self, "id", resources.Id.parse_id(self.resource_id))


class DeployResult(StrEnum):
    """
    The result of a resource's last (finished) deploy. This result may be for an older version than the latest desired state.
    See Compliance for a resource's operational status with respect to its latest desired state.

    NEW: Resource has never been deployed before.
    DEPLOYED: Last resource deployment was successful.
    FAILED: Last resource deployment was unsuccessful.
    SKIPPED: Resource skipped deployment.
    """

    NEW = enum.auto()
    DEPLOYED = enum.auto()
    FAILED = enum.auto()
    SKIPPED = enum.auto()

    @classmethod
    def from_handler_resource_state(cls, handler_resource_state: const.HandlerResourceState) -> "DeployResult":
        match handler_resource_state:
            case const.HandlerResourceState.deployed:
                return DeployResult.DEPLOYED
            case const.HandlerResourceState.skipped | const.HandlerResourceState.skipped_for_dependency:
                return DeployResult.SKIPPED
            case const.HandlerResourceState.failed | const.HandlerResourceState.unavailable:
                return DeployResult.FAILED
            case _ as resource_state:
                raise Exception(f"Unexpected handler_resource_state {resource_state.name}")


class AgentStatus(StrEnum):
    """
    The status of the agent responsible of a given resource.

    STARTED: Agent has been started.
    STOPPING: Agent is stopping.
    STOPPED: Agent has been stopped (previously called PAUSED).
    """

    STARTED = enum.auto()
    STOPPING = enum.auto()
    STOPPED = enum.auto()


class Blocked(StrEnum):
    """
    BLOCKED: The resource will retain its blocked status within this model version. For example: A resource that has unknowns
         or depends on a resource with unknowns.
    NOT_BLOCKED: The resource is not blocked
    TEMPORARILY_BLOCKED: The resource is blocked but may recover within the same version.
        Concretely it is waiting for its dependencies to deploy successfully.
    """

    BLOCKED = enum.auto()
    NOT_BLOCKED = enum.auto()
    TEMPORARILY_BLOCKED = enum.auto()

    def db_value(self: "Blocked") -> "Blocked":
        """
        Convert this blocked status to one appropriate for writing to the database.

        This method exists to work around #8541 until a proper solution can be devised.
        """
        # TODO[#8541]: also persist TEMPORARILY_BLOCKED in database
        if self is Blocked.TEMPORARILY_BLOCKED:
            return Blocked.NOT_BLOCKED
        return self


@dataclass
class ResourceState:
    """
    State of a resource. Consists of multiple independent (mostly) state vectors that make up the final state.

    :param last_deployed: when was this resource last deployed
    """

    # FIXME: review / finalize resource state. Based on draft design in
    #   https://docs.google.com/presentation/d/1F3bFNy2BZtzZgAxQ3Vbvdw7BWI9dq0ty5c3EoLAtUUY/edit#slide=id.g292b508a90d_0_5
    compliance: Compliance
    last_deploy_result: DeployResult
    blocked: Blocked
    last_deployed: datetime.datetime | None

    def is_dirty(self) -> bool:
        """
        Return True iff the status indicates the resource is not up-to-date and is ready to be deployed.
        """
        return self.compliance.is_dirty()

    def copy(self: Self) -> Self:
        """
        Returns a copy of this resource state object.
        """
        return dataclasses.replace(self)


@dataclass(kw_only=True)
class ModelState:
    """
    The state of the model, meaning both resource intent and resource state.

    Invariant: all resources in the current model, and only those, exist in the resources (also undeployable resources)
               and state mappings.
    """

    version: int
    intent: dict["ResourceIdStr", ResourceIntent] = dataclasses.field(default_factory=dict)
    requires: RequiresProvidesMapping = dataclasses.field(default_factory=RequiresProvidesMapping)
    resource_state: dict["ResourceIdStr", ResourceState] = dataclasses.field(default_factory=dict)
    # resources with a known or assumed difference between intent and actual state
    dirty: set["ResourceIdStr"] = dataclasses.field(default_factory=set)
    # group resources by agent to allow efficient triggering of a deploy for a single agent
    resources_by_agent: dict[str, set["ResourceIdStr"]] = dataclasses.field(default_factory=lambda: defaultdict(set))
    # types per agent keeps track of which resource types live on which agent by doing a reference count
    # the dict is agent_name -> resource_type -> resource_count
    types_per_agent: dict[str, dict["ResourceType", int]] = dataclasses.field(
        default_factory=lambda: defaultdict(lambda: defaultdict(lambda: 0))
    )

    @classmethod
    async def create_from_db(
        cls, environment: uuid.UUID, *, connection: asyncpg.connection.Connection
    ) -> Optional["ModelState"]:
        """
        Create a new instance of the ModelState object, by restoring the model state from the database.
        Returns None iff no such (released) version exists (e.g. the scheduler hasn't processed any
        versions in the past or its last seen version has been cleaned up).

        :param environment: The environment the model state belongs to.
        """
        from inmanta import data, resources

        scheduler: Optional[data.Scheduler] = await data.Scheduler.get_one(environment=environment, connection=connection)
        assert scheduler is not None
        last_processed_model_version: Optional[int] = scheduler.last_processed_model_version
        if last_processed_model_version is None:
            return None

        result = ModelState(version=last_processed_model_version)
        resource_records: Sequence[Mapping[str, object]] = (
            await data.Resource.get_resources_for_version_raw_with_persistent_state(
                environment=environment,
                version=last_processed_model_version,
                projection=["resource_id", "attributes", "attribute_hash"],
                projection_persistent=[
                    "is_orphan",
                    "is_undefined",
                    "current_intent_attribute_hash",
                    "last_deployed_attribute_hash",
                    "last_deploy_result",
                    "blocked",
                    "last_success",
                    "last_deploy",
                    "last_produced_events",
                ],
                project_attributes=[
                    "requires",
                    const.RESOURCE_ATTRIBUTE_SEND_EVENTS,
                    const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS,
                ],
                connection=connection,
            )
        )
        if not resource_records:
            configuration_model: Optional[data.ConfigurationModel] = await data.ConfigurationModel.get_one(
                environment=environment, version=last_processed_model_version, released=True, connection=connection
            )
            if configuration_model is None:
                # the version does not exist at all (anymore)
                return None
            # otherwise it's simply an empty version => continue with normal flow
        by_resource_id: Mapping[ResourceIdStr, Mapping[str, object]] = {
            ResourceIdStr(cast(str, r["resource_id"])): r for r in resource_records
        }
        for resource_id, res in by_resource_id.items():
            # Populate state

            compliance_status: Compliance
            last_deployed = cast(datetime.datetime, res["last_deploy"])
            if res["is_orphan"]:
                # it was marked as an orphan by the scheduler when (or sometime before) it read the version we're currently
                # processing => exclude it from the model
                continue
            elif res["is_undefined"]:
                # it was marked as undefined by the scheduler when it read the version we're currently processing
                # (scheduler is only writer)
                compliance_status = Compliance.UNDEFINED
            elif (
                DeployResult[res["last_deploy_result"]] is DeployResult.NEW
                or res["last_deployed_attribute_hash"] is None
                or res["current_intent_attribute_hash"] != res["last_deployed_attribute_hash"]
            ):
                compliance_status = Compliance.HAS_UPDATE
            elif DeployResult[res["last_deploy_result"]] is DeployResult.DEPLOYED:
                compliance_status = Compliance.COMPLIANT
            else:
                compliance_status = Compliance.NON_COMPLIANT

            resource_state = ResourceState(
                compliance=compliance_status,
                last_deploy_result=DeployResult[res["last_deploy_result"]],
                blocked=Blocked[res["blocked"]],
                last_deployed=last_deployed,
            )
            result.resource_state[resource_id] = resource_state

            # Populate resources intent
            resource_intent = ResourceIntent(
                resource_id=resource_id,
                attribute_hash=res["attribute_hash"],
                attributes=json.loads(res["attributes"]),
            )
            result.intent[resource_id] = resource_intent

            # Populate types_per_agent
            result.types_per_agent[resource_intent.id.agent_name][resource_intent.id.entity_type] += 1
            result.resources_by_agent[resource_intent.id.agent_name].add(resource_id)

            # Populate requires
            requires = {resources.Id.parse_id(req).resource_str() for req in res["requires"]}
            result.requires[resource_id] = requires

            # Check whether resource is dirty
            if resource_state.blocked is Blocked.NOT_BLOCKED:
                if resource_state.is_dirty():
                    # Resource is dirty by itself.
                    result.dirty.add(resource_intent.id.resource_str())
                elif res.get(const.RESOURCE_ATTRIBUTE_RECEIVE_EVENTS, True):
                    # Check whether the resource should be deployed because of an outstanding event.
                    last_success = res["last_success"] or const.DATETIME_MIN_UTC
                    for req in requires:
                        req_res = by_resource_id[req]
                        assert req_res is not None
                        last_produced_events = req_res["last_produced_events"]
                        if (
                            last_produced_events is not None
                            and last_produced_events > last_success
                            and req_res.get(const.RESOURCE_ATTRIBUTE_SEND_EVENTS, False)
                        ):
                            result.dirty.add(resource_intent.id.resource_str())
                            break
        return result

    def reset(self) -> None:
        self.version = 0
        self.intent.clear()
        self.requires.clear()
        self.resource_state.clear()
        self.resources_by_agent.clear()
        self.dirty = set()

    def update_resource(
        self,
        resource_intent: ResourceIntent,
        *,
        force_new: bool = False,
        undefined: bool = False,
        known_compliant: bool = False,
        last_deployed: datetime.datetime | None = None,
    ) -> None:
        """
        Register a change of intent for a resource. Registers the new resource intent, as well as its undefined status.
        Does not touch or take into account requires-provides. To update these, call update_requires().

        Sets blocked status for undefined resources, but not vice versa because this depends on transitive properties, which
        requires a full view, including fully updated requires. When you call this method, you must also call
        update_transitive_state() after all direct state and requires have been updated.

        :param force_new: Whether to consider this a new resource, even if we happen to know one with the same id
            (in which case the old one is dropped from the model before registering the new one).
        :param undefined: Whether this resource's intent is undefined, i.e. there's an unknown attribute. Mutually exclusive
            with known_compliant.
        :param known_compliant: Whether this resource is known to be in a good state (compliant and last deploy was successful).
            Useful for restoring previously known state when scheduler is started. Mutually exclusive with undefined.
        :param last_deployed: last deployed time. Only set when restoring previously known state when scheduler is started.
        """
        if undefined and known_compliant:
            raise ValueError("A resource can not be both undefined and compliant")

        resource: ResourceIdStr = resource_intent.resource_id
        compliance_status: Compliance = (
            Compliance.COMPLIANT if known_compliant else Compliance.UNDEFINED if undefined else Compliance.HAS_UPDATE
        )
        # Latest requires are not set yet, transitve blocked status are handled in update_transitive_state
        blocked: Blocked = Blocked.BLOCKED if undefined else Blocked.NOT_BLOCKED

        already_known: bool = resource_intent.resource_id in self.intent
        if force_new and already_known:
            # register this as a new resource, even if we happen to know one with the same id
            self.drop(resource_intent.resource_id)

        if not already_known or force_new:
            # we don't know the resource yet (/ anymore) => create it
            self.resource_state[resource] = ResourceState(
                compliance=compliance_status,
                last_deploy_result=DeployResult.DEPLOYED if known_compliant else DeployResult.NEW,
                blocked=blocked,
                last_deployed=last_deployed,
            )
            if resource not in self.requires:
                self.requires[resource] = set()
            self.types_per_agent[resource_intent.id.agent_name][resource_intent.id.entity_type] += 1
            self.resources_by_agent[resource_intent.id.agent_name].add(resource)
        else:
            # we already know the resource => update relevant fields
            self.resource_state[resource].compliance = compliance_status
            # update deployment result only if we know it's compliant. Otherwise it is kept, representing latest result
            if known_compliant:
                self.resource_state[resource].last_deploy_result = DeployResult.DEPLOYED
            # Override blocked status except if it was marked as blocked before. We can't unset it yet because a resource might
            # still be transitively blocked, which we'll deal with later, see note in docstring.
            # We do however override TEMPORARILY_BLOCKED because we want to give it another chance when it gets an update
            # (in part to progress the resource state away from available).
            if self.resource_state[resource].blocked is not Blocked.BLOCKED:
                self.resource_state[resource].blocked = blocked

        self.intent[resource] = resource_intent
        if not known_compliant and self.resource_state[resource].blocked is Blocked.NOT_BLOCKED:
            self.dirty.add(resource)
        else:
            self.dirty.discard(resource)

    def update_requires(self, resource: "ResourceIdStr", requires: Set["ResourceIdStr"]) -> None:
        """
        Update the requires relation for a resource. Updates the reverse relation accordingly.

        When updating requires, also call update_transitive_state to ensure temporirly_blocked state is also updated
        """
        check_dependencies: bool = self.resource_state[resource].blocked is Blocked.TEMPORARILY_BLOCKED and bool(
            self.requires[resource] - requires
        )
        self.requires[resource] = requires
        # If the resource is blocked temporarily, and we drop at least one of its requirements
        # we check to see if the resource can now be unblocked
        # i.e. all of its dependencies are now compliant with the desired state.
        if check_dependencies and not self.should_skip_for_dependencies(resource):
            self.resource_state[resource].blocked = Blocked.NOT_BLOCKED
            self.dirty.add(resource)

    def drop(self, resource: "ResourceIdStr") -> None:
        """
        Completely remove a resource from the resource state.
        """
        resource_intent: ResourceIntent = self.intent.pop(resource)
        del self.resource_state[resource]
        # stand-alone resources may not be in requires
        with contextlib.suppress(KeyError):
            del self.requires[resource]
        # top-level resources may not be in provides
        with contextlib.suppress(KeyError):
            del self.requires.reverse_mapping()[resource]

        self.types_per_agent[resource_intent.id.agent_name][resource_intent.id.entity_type] -= 1
        if self.types_per_agent[resource_intent.id.agent_name][resource_intent.id.entity_type] == 0:
            del self.types_per_agent[resource_intent.id.agent_name][resource_intent.id.entity_type]
        self.resources_by_agent[resource_intent.id.agent_name].discard(resource)
        if not self.resources_by_agent[resource_intent.id.agent_name]:
            del self.resources_by_agent[resource_intent.id.agent_name]
        self.dirty.discard(resource)

    def should_skip_for_dependencies(self, resource: "ResourceIdStr") -> bool:
        """
        Returns true if this resource satisfies the conditions to skip for dependencies, provided that its handler requested it.
        Concretely, checks if a resource has at least one dependency that was not successful in its last deploy, if any.

        :param resource: The id of the resource to find the dependencies for
        """
        dependencies: Set[ResourceIdStr] = self.requires.get(resource, set())
        return any(
            # Determine based on latest deploy result rather than compliance status, because compliance status is unstable
            # (e.g. HAS_UPDATE).
            # Make sure to not skip on NEW (never deployed) resources, because they will not generate a discovery event.
            self.resource_state[dep_id].last_deploy_result not in (DeployResult.DEPLOYED, DeployResult.NEW)
            for dep_id in dependencies
        )

    def update_transitive_state(
        self,
        *,
        new_undefined: "Set[ResourceIdStr]",
        verify_blocked: "Set[ResourceIdStr]",
        verify_unblocked: "Set[ResourceIdStr]",
    ) -> "tuple[Set[ResourceIdStr], Set[ResourceIdStr]]":
        """

        Update transitive states.
         Assumes transitive states are consistently set, apart from those of the resources passed as parameters.
         In turn this method ensures transitive state consistency for the entire model.

        :param new_undefined: resources that have become undefined.
            These have already moved to the blocked state by update_resource
        :param verify_blocked: resources that may have gotten blocked, e.g. due to added requires.
            If blocked, will propagate the check to its provides even if not in this set.
        :param verify_unblocked: resources that may have gotten unblocked,
            e.g. due to a transition from undefined to defined or due to dropped requires.
            If unblocked, will propagate the check to its provides even if not in this set.

        returns: <unblocked, blocked> resources that have effectively been moved to this state by this method.
          note: this doesn't include the members of new_undefined, as they had already been updated
        """
        # This algorithm moves the blocked/unblocked forward over the provides relation
        # 1. We start from the nodes passed in
        # 2. Resources that have become undefined are first marked over the requires relation.
        #   - This is easy as blocked state with a known root can safely be propagated froward
        # 3. Resource that added a requirement may have become blocked
        #   - We back check one step  (along requires)
        #        (safe because of invariant that all transitive states not passed to this method are set consistently)
        #   - We mark them as blocked if needed.
        #       - Nodes marked as blocked this way are not definitely blocked,
        #         as nodes deeper down the tree may become unblocked later on
        #       - Except if we find an undefined node on the path, in that case, they are definitely blocked
        #   - We forward propagate that
        # 4. Resources that removed a requirements or became defined
        #    are checked one step along the requires relations to see if they are unblocked
        #    - We check back one step (along requires)
        #       (safe because of invariant that all transitive states not passed to this method are set consistently)
        #    - We mark them as unblocked if needed
        #       - Nodes marked as blocked this way are not definitely blocked,
        #         as nodes deeper down the tree may become unblocked later on
        #       - Except if we find an undefined node on the path, in that case, they are definitely blocked
        #    - We forward propagate that

        # To traverse efficiently, we maintain:
        #  - a worklist of definitely blocked nodes
        #  - a set of definitely blocked nodes
        #  - a worklist of potential unblocked nodes

        #  - a cache of blocker nodes. When we find a node B blocking a potentially unblocked node A, we cache this.
        #    If we hit node A again and we are not coming from B, we know it is still blocked.

        out_blocked: set[ResourceIdStr] = set()
        out_unblocked: set[ResourceIdStr] = set()

        # a set of definitely blocked nodes
        is_blocked: set[ResourceIdStr] = set(new_undefined)

        # forward graph
        provides_view: Mapping[ResourceIdStr, Set[ResourceIdStr]] = self.requires.provides_view()

        # 2. forward propagate blocked, start one provides step forward from the updated nodes
        propagate_blocked_work: list[ResourceIdStr] = list(
            itertools.chain.from_iterable(provides_view.get(r, set()) for r in is_blocked)
        )
        while propagate_blocked_work:
            resource_id: ResourceIdStr = propagate_blocked_work.pop()
            if resource_id in is_blocked:
                continue
            # directly down the provides relation from an undefined: definitely blocked,
            # will not be unblocked by stage 4
            is_blocked.add(resource_id)
            if self.resource_state[resource_id].blocked is Blocked.BLOCKED:
                # Resource is already blocked. All provides will be blocked as well.
                continue
            else:
                # Block resource and its provides.
                self._block_resource_transitive(resource_id)
                out_blocked.add(resource_id)
                propagate_blocked_work.extend(provides_view.get(resource_id, set()))

        # 3. back check propagate potential blockers
        known_blockers_cache: dict[ResourceIdStr, ResourceIdStr] = {}
        # Cache used by the `update_blocked_status_*()` method to prevent the expensive check on the blocked
        # status of the requirements of a resource. If the dictionary contains key value pair (X, Y), it
        # indicates that resource Y is a requirement of resource X that was blocked, last we checked
        # This cache is a bit complicated in that
        #  - The unblock check on node A takes us up to n traversals (if A has n requires)
        #  - We may hit A up to n times
        #  - The chance of a cache miss is 1/n (i.e. we come from that edge) (roughly)
        #  - So we reduce from n*n to n time complexity
        # unclear if it is worth it, hard to tell/test

        def update_blocked_status_to_blocked(resource: "ResourceIdStr") -> bool:
            """
            Check if the deployment of the given resource is still blocked and mark it as unblocked if it is.
            check one step over the requires relation, assume all states to hold.
            if they don't we will pass this node again

            :return: True iff the given resource changes state.
            """
            if resource in is_blocked:
                # Is blocked for a know root
                return False

            my_state = self.resource_state[resource]
            if my_state.blocked is Blocked.BLOCKED:
                # The resource is already blocked.
                return False

            blocked_dependency: ResourceIdStr | None = None
            # one blocked requires, evidence we are blocked

            if resource in known_blockers_cache:
                # First check the blocked status of the cached known blocker for improved performance.
                known_blocker: ResourceIdStr = known_blockers_cache[resource]
                if self.resource_state[known_blocker].blocked is Blocked.BLOCKED:
                    blocked_dependency = known_blocker
                else:
                    # Cache is out of date. Clear cache item.
                    del known_blockers_cache[resource]

            if blocked_dependency is None:
                # Perform more expensive call by traversing all requirements of resource.
                blocked_dependency = next(
                    (r for r in self.requires.get(resource, set()) if self.resource_state[r].blocked is Blocked.BLOCKED), None
                )
                if blocked_dependency is not None:
                    # cache the resource that is blocking us
                    known_blockers_cache[resource] = blocked_dependency

            if blocked_dependency is None:
                return False

            if blocked_dependency in is_blocked:
                # We know the root is undefined, hard block
                is_blocked.add(resource)

            # Block resource
            self._block_resource_transitive(resource)
            out_blocked.add(resource)
            return True

        for resource_id in verify_blocked:
            # verify all potential blockers
            blocked = update_blocked_status_to_blocked(resource_id)
            # if we did block it, add it to the work
            if blocked:
                propagate_blocked_work.append(resource_id)

        # Propagate blockers once more
        # We iterate differently, because the parents blocked status is relevant
        while propagate_blocked_work:
            already_blocked: ResourceIdStr = propagate_blocked_work.pop()
            is_hard_block = already_blocked in is_blocked
            for to_be_blocked in provides_view.get(already_blocked, set()):
                if is_hard_block:
                    is_blocked.add(to_be_blocked)
                my_state = self.resource_state[to_be_blocked]
                if my_state.blocked is not Blocked.BLOCKED:
                    continue
                else:
                    self._block_resource_transitive(to_be_blocked)
                    out_blocked.add(to_be_blocked)
                    propagate_blocked_work.append(to_be_blocked)

        # 4. forward propagate the unblocked
        def update_blocked_status(resource: "ResourceIdStr") -> bool:
            """
            Check if the deployment of the given resource is still blocked and mark it as unblocked if it is.
            check one step over the requires relation, assume all states to hold.
            if they don't we will pass this node again

            :return: True iff the given resource was unblocked.
            """
            if resource in is_blocked:
                # Is blocked for a know root
                return False

            my_state = self.resource_state[resource]
            if my_state.blocked is not Blocked.BLOCKED:
                # The resource is already unblocked.
                return False

            if my_state.compliance is Compliance.UNDEFINED:
                # The resource is undefined.
                # Root blocker
                is_blocked.add(resource)
                return False

            if resource in known_blockers_cache:
                # First check the blocked status of the cached known blocker for improved performance.
                known_blocker: "ResourceIdStr" = known_blockers_cache[resource]
                if self.resource_state[known_blocker].blocked is Blocked.BLOCKED:
                    return False
                else:
                    # Cache is out of date. Clear cache item.
                    del known_blockers_cache[resource]

            # Perform more expensive call by traversing all requirements of resource.
            blocked_dependency: "ResourceIdStr" | None = next(
                (r for r in self.requires.get(resource, set()) if self.resource_state[r].blocked is Blocked.BLOCKED), None
            )
            if blocked_dependency:
                # Resource is blocked, because a dependency is blocked.
                known_blockers_cache[resource] = blocked_dependency
                return False

            # Unblock resource
            self._unblock_resource(resource)
            out_unblocked.add(resource)
            return True

        to_verify_work: "list[ResourceIdStr]" = list(verify_unblocked)
        while to_verify_work:
            resource_id = to_verify_work.pop()
            resource_was_unblocked: bool = update_blocked_status(resource_id)
            if resource_was_unblocked:
                to_verify_work.extend(provides_view.get(resource_id, set()))

        return out_unblocked, out_blocked

    def _block_resource_transitive(self, resource: "ResourceIdStr") -> None:
        """
        Mark the given resource as blocked, it's not deployable.

        Only used internally, see update_transitive_state
        """

        self.resource_state[resource].blocked = Blocked.BLOCKED
        self.dirty.discard(resource)

    def _unblock_resource(self, resource: "ResourceIdStr") -> None:
        """
        Mark the given resource as unblocked

        Only used internally, see update_transitive_state
        """
        my_state = self.resource_state[resource]
        my_state.blocked = Blocked.NOT_BLOCKED
        if my_state.compliance in [Compliance.HAS_UPDATE, Compliance.NON_COMPLIANT]:
            self.dirty.add(resource)



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import enum
import functools
import logging
import typing
import uuid

from inmanta.config import *
from inmanta.server.config import db_connection_pool_max_size, db_connection_pool_min_size

LOGGER = logging.getLogger(__name__)

# flake8: noqa: H904
environment: Option[typing.Optional[uuid.UUID]] = Option(
    "config", "environment", None, "The environment this agent or compile belongs to", is_uuid_opt
)

agent_reconnect_delay: Option[int] = Option(
    "config", "agent-reconnect-delay", 5, "Time to wait after a failed heartbeat message. DO NOT SET TO 0 ", is_int
)

server_timeout = Option(
    "config",
    "server-timeout",
    125,
    "Amount of time to wait for a response from the server before we try to reconnect, must be larger than server.agent-hold",
    is_time,
)

agent_deploy_interval: Option[int | str] = Option(
    "config",
    "agent-deploy-interval",
    0,
    "Set the frequency and the granularity of deploy runs (i.e. only trigger a deploy for resources that have a "
    " known divergence with their desired state)."
    " When specified as an integer, this will set the wait time (in seconds) before attempting to redeploy a resource"
    " after an unsuccessful deployment, on a per-resource basis."
    " When specified as a cron-like expression, a global deploy (i.e. for all resources that have a known divergence with their"
    " desired state) will be run following a cron-like time-to-run specification,"
    " interpreted in UTC. The expected format is ``[sec] min hour dom month dow [year]`` (If only 6 values are provided, they"
    " are interpreted as ``min hour dom month dow year``)."
    " A deploy will be requested at the scheduled time."
    " Set this to 0 to disable the scheduled deploy runs.",
    is_time_or_cron,
)

agent_repair_interval = Option(
    "config",
    "agent-repair-interval",
    600,
    "Set the frequency and the granularity of repair runs (i.e. trigger a deploy regardless of the assumed"
    " state of the resource(s))."
    " When specified as an integer, this will set the wait time (in seconds) before re-scheduling a resource for deployment"
    " after the previous deployment has ended, regardless of success or failure, on a per-resource basis."
    " When specified as a cron-like expression, a global repair (i.e. a full deploy for all resources, regardless of their"
    " assumed desired state and regardless of their actual state) will be run following a cron-like"
    " time-to-run specification,"
    " interpreted in UTC. The expected format is `[sec] min hour dom month dow [year]` ( If only 6 values are provided, they"
    " are interpreted as `min hour dom month dow year`)."
    " A repair will be requested at the scheduled time."
    " Setting this to 0 to disable the scheduled repair runs.",
    is_time_or_cron,
)

executor_venv_retention_time: Option[int] = Option(
    "agent",
    "executor-venv-retention-time",
    3600,
    "This is the number of seconds to wait before unused Python virtual environments of an executor are removed from "
    "the inmanta server. Setting this option too low may result in a high load on the Inmanta server. Setting it too high"
    " may result in increased disk usage.",
    # We know that the .inmanta venv status file is touched every minute, so `60` seconds is the lowest default we can use
    is_lower_bounded_int(60),
)


def default_db_pool_min_size() -> int:
    """:inmanta.config:option:`database.connection-pool-min-size` / 10"""
    return int(db_connection_pool_min_size.get() / 10)


def default_db_pool_max_size() -> int:
    """:inmanta.config:option:`database.connection-pool-max-size` / 10"""
    return int(db_connection_pool_max_size.get() / 10)


scheduler_db_connection_pool_min_size: Option[int] = Option(
    "scheduler",
    "db-connection-pool-min-size",
    default_db_pool_min_size,
    "In each environment, the database connection pool will be initialized with this number of connections for "
    "the resource scheduler.",
    is_lower_bounded_int(0),
)
scheduler_db_connection_pool_max_size: Option[int] = Option(
    "scheduler",
    "db-connection-pool-max-size",
    default_db_pool_max_size,
    "In each environment, limit the size of the database connection pool to this number of connections for "
    "the resource scheduler.",
    is_lower_bounded_int(1),
)
scheduler_db_connection_timeout: Option[float] = Option(
    "scheduler",
    "db-connection-timeout",
    60.0,
    "In each environment, set the database connection timeout for interactions of the scheduler with the database"
    " (in seconds).",
    is_float,
)


class AgentExecutorMode(str, enum.Enum):
    threaded = "threaded"
    forking = "forking"


def is_executor_mode(value: str | AgentExecutorMode) -> AgentExecutorMode:
    """threaded | forking"""
    if isinstance(value, AgentExecutorMode):
        return value
    return AgentExecutorMode(value)


agent_executor_cap = Option[int](
    "agent",
    "executor-cap",
    3,
    "Maximum number of concurrent executors to keep per environment, per agent. If this limit is already reached "
    "when creating a new executor, the oldest one will be stopped first.",
    is_lower_bounded_int(1),
)

agent_executor_retention_time = Option[int](
    "agent",
    "executor-retention-time",
    60,
    "Amount of time (in seconds) to wait before cleaning up inactive executors.",
    is_time,
)

agent_cache_cleanup_tick_rate = Option[int](
    "agent",
    "cache-cleanup-tick-rate",
    1,
    "The rate (in seconds) at which the agent will periodically attempt to remove stale entries from the cache when idle.",
    is_time,
)

##############################
# agent_rest_transport
##############################

agent_transport = TransportConfig("agent")



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from collections.abc import Hashable, Iterator, Mapping, MutableMapping, Set
from typing import Generic, Optional, TypeVar

S = TypeVar("S", bound=Hashable)
T = TypeVar("T", bound=Hashable)
# type vars not bound to the class
K = TypeVar("K", bound=Hashable)
V = TypeVar("V", bound=Hashable)


class BidirectionalManyMapping(MutableMapping[S, Set[T]], Generic[S, T]):
    """
    A mutable bidirectional mapping many-to-many mapping between two value domains. All operations uphold the bidirectional
    invariant.

    e.g. if a value s maps to t1 and t2, then both t1 and t2 will include s in the reverse mapping. This property holds in both
    directions.

    While this is a mutable mapping, mutations on the underlying sets are not supported.
    """

    def __init__(self, mapping: Optional[Mapping[S, Set[T]]] = None) -> None:
        self._primary: dict[S, set[T]] = {}
        self._reverse: dict[T, set[S]] = {}
        if mapping is not None:
            for key, values in mapping.items():
                self[key] = values

    @staticmethod
    def _set(primary: dict[K, set[V]], reverse: dict[V, set[K]], key: K, values: Set[V]) -> None:
        """
        Set the key-value pair on the primary dict and update the reverse mapping to uphold the bidirectionality.

        :param primary: The dict to consider as primary for this set operation, i.e. the one to set the key-value pair on.
        :param reverse: The dict to consider as secondary for this set operation, i.e. the one to apply the reverse key-value
            mapping on (may include deleting edges that were dropped from the primary mapping).
        """
        current: Set[V] = primary.get(key, set())
        new: Set[V] = values - current
        missing: Set[V] = current - values

        # update primary
        primary[key] = set(values)

        # update reverse
        v: V
        for v in new:
            if v not in reverse:
                reverse[v] = set()
            reverse[v].add(key)
        for v in missing:
            reverse[v].remove(key)

    # MutableMapping interface

    def __getitem__(self, key: S) -> Set[T]:
        return self._primary.__getitem__(key)

    def __setitem__(self, key: S, value: Set[T]) -> None:
        self._set(self._primary, self._reverse, key, value)

    def __delitem__(self, key: S) -> None:
        """
        Delete item from this end of the mapping.
        """
        if key not in self._primary:
            raise KeyError(key)
        # trim reverse mapping
        self[key] = set()
        # delete from primary mapping
        del self._primary[key]

    def __iter__(self) -> Iterator[S]:
        return iter(self._primary)

    def __len__(self) -> int:
        return len(self._primary)

    def __repr__(self) -> str:
        # show both directions because empty sets are only visible on one end
        return f"BidirectionalManyMapping(primary={self._primary!r}, reverse={self._reverse!r})"

    def __str__(self) -> str:
        return str(self._primary)

    # Methods for reverse access

    def reverse_mapping(self) -> "BidirectionalManyMapping[T, S]":
        """
        Return a BidirectionalManyMapping coupled to this one to represent the reverse mapping. The instance remains coupled
        with this one, meaning that changes to one are reflected to the other, i.e. they're mutable views on both sides of the
        same bidirectional mapping.

        This method is symmetric, i.e. `self.reverse_mapping().reverse_mapping() is self`
        """
        return _BidirectionalManyToManyMappingReverse(self)

    def get_reverse(self, key: T, default: Optional[Set[S]] = None) -> Optional[Set[S]]:
        """
        Return the values associated with the given key in the reverse mapping.

        Equivalent to `self.reverse_mapping()[key].get()`
        """
        return self._reverse.get(key, default)

    def set_reverse(self, key: T, values: Set[S]) -> None:
        """
        Set new values for the key in the reverse mapping. Old values are completely replaced and the primary mapping is
        updated accordingly.

        Equivalent to `self.reverse_mapping()[key] = values`
        """
        self._set(self._reverse, self._primary, key, values)


class _BidirectionalManyToManyMappingReverse(BidirectionalManyMapping[T, S], Generic[S, T]):
    def __init__(self, base: BidirectionalManyMapping[S, T]) -> None:
        self._base = base
        self._primary = self._base._reverse
        self._reverse = self._base._primary

    def reverse_mapping(self) -> BidirectionalManyMapping[S, T]:
        return self._base



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

# This file defines named type definition for the Inmanta code base

import builtins
import datetime
import uuid
from collections.abc import Coroutine, Generator, Mapping, Sequence
from typing import TYPE_CHECKING, Any, Callable, ClassVar, NewType, Optional, Union

import pydantic
import typing_inspect

from inmanta.stable_api import stable_api

if TYPE_CHECKING:
    # Include imports from other modules here and use the quoted annotation in the definition to prevent import loops
    from inmanta.protocol.common import ReturnValue  # noqa: F401


# Typing of dataclass.* methods relies entirely on the definition in typeshed that only exists during typechecking.
# This ensures that our code works during typechecking and at runtime.
if not TYPE_CHECKING:
    DataclassProtocol = object
else:
    import _typeshed

    DataclassProtocol = _typeshed.DataclassInstance


def api_boundary_datetime_normalizer(value: datetime.datetime) -> datetime.datetime:
    if value.tzinfo is None:
        return value.replace(tzinfo=datetime.timezone.utc)
    else:
        return value


@stable_api
class DateTimeNormalizerModel(pydantic.BaseModel):
    """
    A model that normalizes all datetime values to be timezone aware. Assumes that all naive timestamps represent UTC times.
    """

    @pydantic.field_validator("*", mode="after")
    @classmethod
    def validator_timezone_aware_timestamps(cls: type, value: object) -> object:
        """
        Ensure that all datetime times are timezone aware.
        """
        if isinstance(value, datetime.datetime):
            return api_boundary_datetime_normalizer(value)
        else:
            return value


@stable_api
class BaseModel(DateTimeNormalizerModel):
    """
    Base class for all data objects in Inmanta
    """

    # Populate models with the value property of enums, rather than the raw enum.
    # This is useful to serialise model.dict() later
    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(use_enum_values=True)


# kept for backwards compatibility
StrictNonIntBool = pydantic.StrictBool


type AsyncioGenerator[R] = Generator[object, None, R]
"""
Asyncio-compatible generator as returned from a sync function, e.g. __await__.
"""
type AsyncioCoroutine[R] = Coroutine[object, None, R]
"""
Coroutine for use with asyncio, where we don't care about yield and send types.
"""


def issubclass(sub: type, super: Union[type, tuple[type, ...]]) -> bool:
    """
    Alternative issubclass implementation that interpretes instances of NewType for the first argument as their super type.
    """
    if typing_inspect.is_new_type(sub):
        return issubclass(sub.__supertype__, super)
    return builtins.issubclass(sub, super)


PrimitiveTypes = Optional[uuid.UUID | bool | int | float | datetime.datetime | str]
type SimpleTypes = BaseModel | PrimitiveTypes

JsonType = dict[str, Any]
ReturnTupple = tuple[int, Optional[JsonType]]
type StrictJson = dict[str, StrictJson] | list[StrictJson] | str | int | float | bool | None

type StrMapping[T] = Mapping[str, T] | Mapping[ResourceIdStr, T] | Mapping[ResourceVersionIdStr, T]

type SinglePageTypes = SimpleTypes | StrMapping[ArgumentTypes]
# only simple types allowed within list args, not dicts or lists.
# Typed as Sequence for necessity (covariance), though runtime checks and method overloads require list in practice.
# This is an unfortunate limitation of the Python type system, related to str being a Sequence, among other things.
# Luckily, list is the more conventional return type for method annotations, so as long as that convention is followed,
# it should not cause any trouble. And if it does after all, the only consequence will be that paging will not be supported
# through the Python client.
type PageableTypes = Sequence[SimpleTypes]

type ArgumentTypes = SinglePageTypes | PageableTypes
type ReturnTypes = SinglePageTypes | PageableTypes

type MethodReturn = ReturnTypes | "ReturnValue[ReturnTypes]"
type MethodType = Callable[..., MethodReturn]

type Apireturn = int | ReturnTupple | "ReturnValue[ReturnTypes]" | "ReturnValue[None]" | ReturnTypes
type Warnings = Optional[list[str]]
type HandlerType = Callable[..., AsyncioCoroutine[Apireturn]]


ResourceVersionIdStr = NewType("ResourceVersionIdStr", str)  # Part of the stable API
"""
    The resource id with the version included.
"""

ResourceIdStr = NewType("ResourceIdStr", str)  # Part of the stable API
"""
    The resource id without the version
"""

ResourceType = NewType("ResourceType", str)
"""
    The type of the resource
"""



"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import logging
import os

from click import testing

from inmanta import data
from inmanta.db.util import PGRestore
from inmanta.server.bootloader import InmantaBootloader
from inmanta.user_setup import cmd, get_connection_pool

logger = logging.getLogger(__name__)


class CLI_user_setup:
    async def run(self, run_locally, username, password, *args, **kwargs):
        # set column width very wide so lines are not wrapped
        os.environ["COLUMNS"] = "1000"
        runner = testing.CliRunner()

        def invoke():
            return runner.invoke(cli=cmd, input=f"{run_locally}\n{username}\n{password}")

        result = await asyncio.get_event_loop().run_in_executor(None, invoke)
        # reset to default again
        del os.environ["COLUMNS"]
        return result


def setup_config(tmpdir, postgres_db, database_name):
    """
    set up the needed config to use usersetup
    """
    dot_inmanta_cfg_file = os.path.join(tmpdir, ".inmanta.cfg")
    with open(dot_inmanta_cfg_file, "w", encoding="utf-8") as f:
        f.write(
            f"""
    [server]
    auth=true
    auth_method=database

    [auth_jwt_default]
    algorithm=HS256
    sign=true
    client_types=agent,compiler,api
    key=eciwliGyqECVmXtIkNpfVrtBLutZiITZKSKYhogeHMM
    expire=0
    issuer=https://localhost:8888/
    audience=https://localhost:8888/

    [database]
    name={database_name}
    host=localhost
    port={str(postgres_db.port)}
    username={postgres_db.user}
    password={postgres_db.password}
    connection_timeout=3
            """
        )
    os.chdir(tmpdir)


async def test_user_setup(tmpdir, server_pre_start, postgres_db, database_name, hard_clean_db, hard_clean_db_post):
    ibl = InmantaBootloader(configure_logging=True)
    # we need the server to start so that all the migrations scripts are applied, but the server needs
    # to be shut down afterwards, otherwise the call to get_connection_pool() will result in an exception saying
    # that the connection pool is already set in the database layer.
    await ibl.start()
    await ibl.stop(timeout=20)

    setup_config(tmpdir, postgres_db, database_name)
    cli = CLI_user_setup()
    result = await cli.run("yes", "new_user", "pw")
    assert result.exit_code == 1
    assert result.stderr == "Error: the password should be at least 8 characters long\n"

    result = await cli.run("yes", "new_user", "password")
    assert result.exit_code == 0
    try:
        # Because the setup command calls data.disconnect(), we cannot use the init_dataclasses_and_load_schema fixture here.
        # After calling into cli.run(), the connection to the database, which was setup by the init_dataclasses_and_load_schema
        # fixture, will be no longer active.
        connection = await get_connection_pool()
        users = await data.User.get_list()
        assert len(users) == 1
        assert users[0].username == "new_user"

    finally:
        if connection is not None:
            await data.disconnect()


async def test_user_setup_empty_username(
    tmpdir, server_pre_start, postgres_db, database_name, hard_clean_db, hard_clean_db_post
):
    """test that if no username is provided to the user setup tool, the username will default to admin"""
    ibl = InmantaBootloader(configure_logging=True)
    # we need the server to start so that all the migrations scripts are applied, but the server needs
    # to be shut down afterwards, otherwise the call to get_connection_pool() will result in an exception saying
    # that the connection pool is already set in the database layer.
    await ibl.start()
    await ibl.stop(timeout=20)

    setup_config(tmpdir, postgres_db, database_name)
    cli = CLI_user_setup()

    result = await cli.run("yes", "", "password")
    assert result.exit_code == 0
    try:
        # Because the setup command calls data.disconnect(), we cannot use the init_dataclasses_and_load_schema fixture here.
        # After calling into cli.run(), the connection to the database, which was setup by the init_dataclasses_and_load_schema
        # fixture, will be no longer active.
        connection = await get_connection_pool()
        users = await data.User.get_list()
        assert len(users) == 1
        assert users[0].username == "admin"

    finally:
        if connection is not None:
            await data.disconnect()


async def test_user_setup_schema_outdated(
    tmpdir, postgres_db, database_name, postgresql_client, hard_clean_db, hard_clean_db_post
):
    setup_config(tmpdir, postgres_db, database_name)

    dump_path = os.path.join(os.path.dirname(__file__), "db/migration_tests/dumps/v202211230.sql")
    with open(dump_path) as fh:
        await PGRestore(fh.readlines(), postgresql_client).run()

    cli = CLI_user_setup()
    result = await cli.run("yes", "new_user", "password")
    assert result.exit_code == 1
    assert (
        result.stderr == "Error: The version of the database is out of date: start the server"
        " to upgrade the database schema to the lastest version.\n"
    )
